#!/usr/bin/env python
"""
Installs FiftyOne.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import os
from pkg_resources import DistributionNotFound, get_distribution
import re
from setuptools import setup, find_packages


VERSION = "0.18.1"


def get_version():
    if "RELEASE_VERSION" in os.environ:
        version = os.environ["RELEASE_VERSION"]
        if not version.startswith(VERSION):
            raise ValueError(
                "Release version does not match version: %s and %s"
                % (version, VERSION)
            )
        return version

    return VERSION


INSTALL_REQUIRES = [
    # third-party packages
    "aiofiles",
    "argcomplete",
    "boto3",
    "dacite>=1.6.0",
    "Deprecated",
    "eventlet",
    "future",
    "hypercorn>=0.13.2",
    "Jinja2>=3",
    "kaleido",
    "matplotlib",
    "mongoengine==0.24.2",
    "motor>=2.3",
    "ndjson",
    "numpy",
    "packaging",
    "pandas",
    "Pillow>=6.2",
    "plotly>=4.14",
    "pprintpp",
    "psutil",
    "pymongo>=3.11",
    "pytz",
    "PyYAML",
    "retrying",
    "scikit-learn",
    "scikit-image",
    "setuptools",
    "sseclient-py>=1.7.2,<2",
    "sse-starlette>=0.10.3,<1",
    "starlette==0.20.4",
    "strawberry-graphql==0.138.1",
    "tabulate",
    "xmltodict",
    "universal-analytics-python3>=1.0.1,<2",
    # internal packages
    "fiftyone-brain>=0.9.2,<0.10",
    "fiftyone-db>=0.4,<0.5",
    "voxel51-eta>=0.8.1,<0.9",
]


CHOOSE_INSTALL_REQUIRES = [
    (
        (
            "opencv-python",
            "opencv-contrib-python",
            "opencv-contrib-python-headless",
        ),
        "opencv-python-headless",
    )
]


def choose_requirement(mains, secondary):
    chosen = secondary
    for main in mains:
        try:
            name = re.split(r"[!<>=]", main)[0]
            get_distribution(name)
            chosen = main
            break
        except DistributionNotFound:
            pass

    return str(chosen)


def get_install_requirements(install_requires, choose_install_requires):
    for mains, secondary in choose_install_requires:
        install_requires.append(choose_requirement(mains, secondary))

    return install_requires


EXTRAS_REQUIREMENTS = {"desktop": ["fiftyone-desktop>=0.24,<0.25"]}


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name="fiftyone",
    version=get_version(),
    description=(
        "FiftyOne: the open-source tool for building high-quality datasets "
        "and computer vision models"
    ),
    author="Voxel51, Inc.",
    author_email="info@voxel51.com",
    url="https://github.com/voxel51/fiftyone",
    extras_require=EXTRAS_REQUIREMENTS,
    license="Apache",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(
        exclude=["app", "eta", "package", "requirements", "tests", "tools"]
    )
    + ["fiftyone.recipes", "fiftyone.tutorials"],
    package_dir={
        "fiftyone.recipes": "docs/source/recipes",
        "fiftyone.tutorials": "docs/source/tutorials",
    },
    install_requires=get_install_requirements(
        INSTALL_REQUIRES, CHOOSE_INSTALL_REQUIRES
    ),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Image Processing",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    entry_points={"console_scripts": ["fiftyone=fiftyone.core.cli:main"]},
    python_requires=">=3.7",
)



"""
FiftyOne Server samples pagination

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import asyncio
from dacite import Config, from_dict
import strawberry as gql
import typing as t


import fiftyone.core.clips as focl
from fiftyone.core.collections import SampleCollection
from fiftyone.core.expressions import ViewField as F
import fiftyone.core.media as fom
import fiftyone.core.odm as foo
from fiftyone.server.filters import SampleFilter

import fiftyone.server.metadata as fosm
from fiftyone.server.paginator import Connection, Edge, PageInfo
import fiftyone.server.view as fosv
from fiftyone.server.scalars import BSON, JSON, BSONArray


@gql.type
class MediaURL:
    field: str
    url: t.Optional[str]


@gql.interface
class Sample:
    id: gql.ID
    sample: JSON
    urls: t.List[MediaURL]


@gql.type
class ImageSample(Sample):
    aspect_ratio: float


@gql.type
class PointCloudSample(Sample):
    pass


@gql.type
class VideoSample(Sample):
    aspect_ratio: float
    frame_rate: float


SampleItem = gql.union(
    "SampleItem", types=(ImageSample, PointCloudSample, VideoSample)
)

MEDIA_TYPES = {
    fom.IMAGE: ImageSample,
    fom.POINT_CLOUD: PointCloudSample,
    fom.VIDEO: VideoSample,
}


async def paginate_samples(
    dataset: str,
    stages: BSONArray,
    filters: BSON,
    first: int,
    after: t.Optional[str] = None,
    extended_stages: t.Optional[BSON] = None,
    sample_filter: t.Optional[SampleFilter] = None,
) -> Connection[t.Union[ImageSample, VideoSample], str]:
    view = fosv.get_view(
        dataset,
        stages=stages,
        filters=filters,
        count_label_tags=True,
        extended_stages=extended_stages,
        sort=True,
        only_matches=True,
        sample_filter=sample_filter,
    )

    root_view = fosv.get_view(
        dataset,
        stages=stages,
    )

    media = view.media_type
    if media == fom.MIXED:
        media = root_view.group_media_types[root_view.default_group_slice]

    if media == fom.GROUP:
        media = view.group_media_types[view.group_slice]

    if after is None:
        after = "-1"

    view = view.skip(int(after) + 1)

    samples = await foo.aggregate(
        foo.get_async_db_conn()[view._dataset._sample_collection_name],
        view._pipeline(
            attach_frames=True,
            detach_frames=False,
            manual_group_select=sample_filter
            and sample_filter.group
            and (sample_filter.group.id and not sample_filter.group.slice),
            support=[1, 1],
        ),
    ).to_list(first + 1)

    more = False
    if len(samples) > first:
        samples = samples[:first]
        more = True

    metadata_cache = {}
    url_cache = {}
    nodes = await asyncio.gather(
        *[
            _create_sample_item(view, sample, metadata_cache, url_cache)
            for sample in samples
        ]
    )

    edges = []
    for idx, node in enumerate(nodes):
        edges.append(
            Edge(
                node=node,
                cursor=str(idx + int(after) + 1),
            )
        )

    return Connection(
        page_info=PageInfo(
            has_previous_page=False,
            has_next_page=more,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if len(edges) > 1 else None,
        ),
        edges=edges,
    )


async def _create_sample_item(
    dataset: SampleCollection,
    sample: t.Dict,
    metadata_cache: t.Dict[str, t.Dict],
    url_cache: t.Dict[str, str],
) -> SampleItem:
    media_type = fom.get_media_type(sample["filepath"])

    if media_type == fom.IMAGE:
        cls = ImageSample
    elif media_type == fom.VIDEO:
        cls = VideoSample
    elif media_type == fom.POINT_CLOUD:
        cls = PointCloudSample
    else:
        raise ValueError(f"unknown media type '{media_type}'")

    metadata = await fosm.get_metadata(
        dataset, sample, media_type, metadata_cache, url_cache
    )

    return from_dict(
        cls,
        {"id": sample["_id"], "sample": sample, **metadata},
        Config(check_types=False),
    )



"""
FiftyOne singleton implementations.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
import weakref


class DatasetSingleton(type):
    """Singleton metaclass for :class:`fiftyone.core.dataset.Dataset`.

    Datasets are singletons keyed by the dataset's ``name``.

    Note that new :class:`fiftyone.core.dataset.Dataset` instances are always
    created if the ``_create == True``.
    """

    def __new__(metacls, *args, **kwargs):
        cls = super().__new__(metacls, *args, **kwargs)
        cls._instances = weakref.WeakValueDictionary()
        return cls

    def __call__(cls, name=None, _create=True, *args, **kwargs):
        if (
            _create
            or name not in cls._instances
            or cls._instances[name].deleted
        ):
            instance = cls.__new__(cls)
            instance.__init__(name=name, _create=_create, *args, **kwargs)
            name = instance.name  # `__init__` may have changed `name`
            cls._instances[name] = instance
        else:
            instance = cls._instances[name]
            instance._update_last_loaded_at()

        return instance


class DocumentSingleton(type):
    """Singleton metaclass interface for
    :class:`fiftyone.core.document.Document` subclasses.

    The methods declared by this interface are used by the
    :class:`fiftyone.core.document.Document` class to manage all instances
    of a class that implements this type.
    """

    def _register_instance(cls, obj):
        """Registers the given instance in the weakref dictionary.

        Args:
            obj: a :class:`Document`
        """
        raise NotImplementedError(
            "subclass must implement _register_instance()"
        )

    def _get_instance(cls, doc):
        """Retrieves the :class:`fiftyone.core.document.Document` instance for
        the given document, if one exists.

        Args:
            doc: a :class:`fiftyone.core.odm.document.Document`

        Returns:
            a :class:`fiftyone.core.document.Document` instance, or None
        """
        raise NotImplementedError("subclass must implement _get_instance()")

    def _reload_instance(cls, obj):
        """Reloads the backing document for the given instance (or view), if a
        reference exists to it.

        Args:
            obj: a :class:`Document` or :class:`DocumentView`
        """
        raise NotImplementedError("subclass must implement _reload_instance()")


class SampleSingleton(DocumentSingleton):
    """Singleton metaclass for :class:`fiftyone.core.sample.Sample`.

    This metaclass maintains a weakref dictionary of all in-memory
    :class:`fiftyone.core.sample.Sample` instances keyed by
    ``[collection name][sample ID]``.
    """

    def __new__(metacls, *args, **kwargs):
        cls = super().__new__(metacls, *args, **kwargs)

        # Instances are keyed by [collection name][sample ID]
        cls._instances = defaultdict(weakref.WeakValueDictionary)

        return cls

    def _register_instance(cls, obj):
        cls._instances[obj._doc.collection_name][obj.id] = obj

    def _get_instance(cls, doc):
        try:
            return cls._instances[doc.collection_name][str(doc.id)]
        except KeyError:
            return None

    def _reload_instance(cls, obj):
        # pylint: disable=no-value-for-parameter
        cls._reload_doc(obj._doc.collection_name, obj.id)

    def _rename_fields(cls, collection_name, field_names, new_field_names):
        """Renames the field on all in-memory samples in the collection."""
        if collection_name not in cls._instances:
            return

        for sample in cls._instances[collection_name].values():
            data = sample._doc._data
            for field_name, new_field_name in zip(
                field_names, new_field_names
            ):
                data[new_field_name] = data.pop(field_name, None)

    def _clear_fields(cls, collection_name, field_names):
        """Clears the values for the given fields (i.e., sets them to None)
        on all in-memory samples in the collection.
        """
        if collection_name not in cls._instances:
            return

        for sample in cls._instances[collection_name].values():
            for field_name in field_names:
                sample._doc._data[field_name] = None

    def _purge_fields(cls, collection_name, field_names):
        """Removes the fields from all in-memory samples in the collection."""
        if collection_name not in cls._instances:
            return

        for sample in cls._instances[collection_name].values():
            for field_name in field_names:
                sample._doc._data.pop(field_name, None)

    def _reload_doc(cls, collection_name, sample_id, hard=False):
        """Reloads the backing document for the given sample if it is
        in-memory.
        """
        if collection_name not in cls._instances:
            return

        sample = cls._instances[collection_name].get(sample_id, None)
        if sample is not None:
            sample.reload(hard=hard)

    def _reload_docs(cls, collection_name, sample_ids=None, hard=False):
        """Reloads the backing documents for in-memory samples in the
        collection.

        If ``sample_ids`` are provided, only those samples are reloaded.
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]

        if sample_ids is not None:
            sample_ids = set(sample_ids)
            for sample in samples.values():
                if sample.id in sample_ids:
                    sample.reload(hard=hard)
        else:
            for sample in samples.values():
                sample.reload(hard=hard)

    def _sync_docs(cls, collection_name, sample_ids, hard=False):
        """Syncs the backing documents for all in-memory samples in the
        collection according to the following rules:

        -   Documents whose IDs are in ``sample_ids`` are reloaded
        -   Documents whose IDs are not in ``sample_ids`` are reset
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]

        sample_ids = set(sample_ids)
        reset_ids = set()
        for sample in samples.values():
            if sample.id in sample_ids:
                sample.reload(hard=hard)
            else:
                reset_ids.add(sample.id)
                sample._reset_backing_doc()

        for sample_id in reset_ids:
            samples.pop(sample_id, None)

    def _reset_docs(cls, collection_name, sample_ids=None):
        """Resets the backing documents for in-memory samples in the
        collection.

        Reset samples will no longer belong to their parent dataset.

        If ``sample_ids`` are provided, only those samples are reset.
        """
        if collection_name not in cls._instances:
            return

        if sample_ids is not None:
            samples = cls._instances[collection_name]
            for sample_id in sample_ids:
                sample = samples.pop(sample_id, None)
                if sample is not None:
                    sample._reset_backing_doc()
        else:
            samples = cls._instances.pop(collection_name)
            for sample in samples.values():
                sample._reset_backing_doc()


class FrameSingleton(DocumentSingleton):
    """Singleton metaclass for :class:`fiftyone.core.frame.Frame`.

    This metaclass maintains a weakref dictionary of all in-memory
    :class:`fiftyone.core.frame.Frame` instances keyed by
    ``[collection name][sample ID][frame number]``.
    """

    def __new__(metacls, *args, **kwargs):
        cls = super().__new__(metacls, *args, **kwargs)

        # Instances are keyed by [collection name][sample ID][frame number]
        cls._instances = defaultdict(
            lambda: defaultdict(weakref.WeakValueDictionary)
        )

        return cls

    def _register_instance(cls, obj):
        cls._instances[obj._doc.collection_name][obj.sample_id][
            obj.frame_number
        ] = obj

    def _get_instance(cls, doc):
        try:
            return cls._instances[doc.collection_name][str(doc._sample_id)][
                doc.frame_number
            ]
        except KeyError:
            return None

    def _reload_instance(cls, obj):
        # pylint: disable=no-value-for-parameter
        cls._reload_doc(
            obj._doc.collection_name, obj.sample_id, obj.frame_number
        )

    def _get_instances(cls, collection_name, sample_id):
        """Returns a frame number -> Frame dict containing all in-memory frame
        instances for the specified sample.
        """
        return dict(cls._instances.get(collection_name, {}).get(sample_id, {}))

    def _rename_fields(cls, collection_name, field_names, new_field_names):
        """Renames the field on all in-memory frames in the collection."""
        if collection_name not in cls._instances:
            return

        for frames in cls._instances[collection_name].values():
            for frame in frames.values():
                data = frame._doc._data
                for field_name, new_field_name in zip(
                    field_names, new_field_names
                ):
                    data[new_field_name] = data.pop(field_name, None)

    def _clear_fields(cls, collection_name, field_names):
        """Clears the values for the given fields (i.e., sets them to None)
        on all in-memory frames in the collection.
        """
        if collection_name not in cls._instances:
            return

        for frames in cls._instances[collection_name].values():
            for frame in frames.values():
                for field_name in field_names:
                    frame._doc._data[field_name] = None

    def _purge_fields(cls, collection_name, field_names):
        """Removes the fields from all in-memory frames in the collection."""
        if collection_name not in cls._instances:
            return

        for frames in cls._instances[collection_name].values():
            for frame in frames.values():
                for field_name in field_names:
                    frame._doc._data.pop(field_name, None)

    def _reload_doc(cls, collection_name, sample_id, frame_number, hard=False):
        """Reloads the backing document for the given frame if it is in-memory."""
        if collection_name not in cls._instances:
            return

        frames = cls._instances[collection_name].get(sample_id, {})
        frame = frames.get(frame_number, None)
        if frame is not None:
            frame.reload(hard=hard)

    def _sync_docs_for_sample(
        cls, collection_name, sample_id, frame_numbers, hard=False
    ):
        """Syncs the backing documents for all in-memory frames attached to the
        specified sample according to the following rules:

        -   Frames whose frame numbers are in ``frame_numbers`` are reloaded
        -   Frames whose frame numbers are not in ``frame_numbers`` are reset
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]
        frames = samples.get(sample_id, {})

        if not frames:
            return

        if callable(frame_numbers):
            frame_numbers = frame_numbers()

        frame_numbers = set(frame_numbers)
        reset_fns = set()

        for frame_number, frame in frames.items():
            if frame_number in frame_numbers:
                frame.reload(hard=hard)
            else:
                reset_fns.add(frame_number)
                frame._reset_backing_doc()

        for frame_number in reset_fns:
            frames.pop(frame_number)

    def _sync_docs(cls, collection_name, sample_ids, hard=False):
        """Syncs the backing documents for all in-memory frames according to
        the following rules:

        -   Frames attached to samples whose IDs are in ``sample_ids`` are
            reloaded
        -   Frames attached to samples whose IDs are not in ``sample_ids`` are
            reset
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]

        sample_ids = set(sample_ids)
        reset_ids = set()
        for sample_id, frames in samples.items():
            if sample_id in sample_ids:
                for frame in frames.values():
                    frame.reload(hard=hard)
            else:
                reset_ids.add(sample_id)
                for frame in frames.values():
                    frame._reset_backing_doc()

        for sample_id in reset_ids:
            frames = samples.pop(sample_id)

    def _reload_docs(cls, collection_name, sample_ids=None, hard=False):
        """Reloads the backing documents for in-memory frames in the
        collection.

        If ``sample_ids`` are provided, only frames attached to samples with
        these IDs are reloaded.
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]

        if sample_ids is not None:
            sample_ids = set(sample_ids)
            for sample_id in sample_ids:
                frames = samples.get(sample_id, {})
                for frame in frames.values():
                    frame.reload(hard=hard)
        else:
            for frames in samples.values():
                for frame in frames.values():
                    frame.reload(hard=hard)

    def _reset_docs(cls, collection_name, sample_ids=None):
        """Resets the backing documents for in-memory frames in the collection.

        If ``sample_ids`` are provided, only frames attached to samples with
        the given sample IDs are reset.
        """
        if collection_name not in cls._instances:
            return

        if sample_ids is not None:
            samples = cls._instances[collection_name]
            for sample_id in sample_ids:
                frames = samples.pop(sample_id, {})
                for frame in frames.values():
                    frame._reset_backing_doc()
        else:
            samples = cls._instances.pop(collection_name)
            for frames in samples.values():
                for frame in frames.values():
                    frame._reset_backing_doc()

    def _reset_docs_for_sample(
        cls, collection_name, sample_id, frame_numbers, keep=False
    ):
        """Resets the backing documents for all in-memory frames with the given
        frame numbers attached to the specified sample.

        When ``keep=True``, all frames whose frame numbers are **not** in
        ``frame_numbers`` are reset instead.
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]
        frames = samples.get(sample_id, {})

        if not frames:
            return

        if callable(frame_numbers):
            frame_numbers = frame_numbers()

        frame_numbers = set(frame_numbers)
        reset_fns = set()

        if keep:
            for frame_number, frame in frames.items():
                if frame_number not in frame_numbers:
                    reset_fns.add(frame_number)
                    frame._reset_backing_doc()
        else:
            for frame_number, frame in frames.items():
                if frame_number in frame_numbers:
                    reset_fns.add(frame_number)
                    frame._reset_backing_doc()

        for frame_number in reset_fns:
            frames.pop(frame_number)

    def _reset_docs_by_frame_id(cls, collection_name, frame_ids, keep=False):
        """Resets the backing documents for all in-memory frames with the given
        frame IDs.

        When ``keep=True``, all frames whose IDs are **not** in ``frame_ids``
        are reset instead.
        """
        if collection_name not in cls._instances:
            return

        samples = cls._instances[collection_name]

        frame_ids = set(frame_ids)
        reset = []

        if keep:
            for sample_id, frames in samples.items():
                for fn, frame in frames.items():
                    if frame.id not in frame_ids:
                        frame._reset_backing_doc()
                        reset.append((sample_id, fn))
        else:
            for sample_id, frames in samples.items():
                for fn, frame in frames.items():
                    if frame.id in frame_ids:
                        frame._reset_backing_doc()
                        reset.append((sample_id, fn))

        for sample_id, fn in reset:
            samples[sample_id].pop(fn)



"""
Patches views.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from copy import deepcopy

from bson import ObjectId

import eta.core.utils as etau

import fiftyone.core.aggregations as foa
import fiftyone.core.dataset as fod
import fiftyone.core.fields as fof
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.odm as foo
import fiftyone.core.sample as fos
import fiftyone.core.validation as fova
import fiftyone.core.view as fov


_PATCHES_TYPES = (fol.Detections, fol.Polylines)
_NO_MATCH_ID = ""


class _PatchView(fos.SampleView):
    @property
    def _sample_id(self):
        return ObjectId(self._doc.sample_id)

    @property
    def _frame_id(self):
        return ObjectId(self._doc.frame_id)

    def _save(self, deferred=False):
        if deferred:
            raise NotImplementedError(
                "Patches views do not support save contexts"
            )

        super()._save(deferred=deferred)
        self._view._sync_source_sample(self)


class PatchView(_PatchView):
    """A patch in a :class:`PatchesView`.

    :class:`PatchView` instances should not be created manually; they are
    generated by iterating over :class:`PatchesView` instances.

    Args:
        doc: a :class:`fiftyone.core.odm.DatasetSampleDocument`
        view: the :class:`PatchesView` that the patch belongs to
        selected_fields (None): a set of field names that this view is
            restricted to
        excluded_fields (None): a set of field names that are excluded from
            this view
        filtered_fields (None): a set of field names of list fields that are
            filtered in this view
    """

    pass


class EvaluationPatchView(_PatchView):
    """A patch in an :class:`EvaluationPatchesView`.

    :class:`EvaluationPatchView` instances should not be created manually; they
    are generated by iterating over :class:`EvaluationPatchesView` instances.

    Args:
        doc: a :class:`fiftyone.core.odm.DatasetSampleDocument`
        view: the :class:`EvaluationPatchesView` that the patch belongs to
        selected_fields (None): a set of field names that this view is
            restricted to
        excluded_fields (None): a set of field names that are excluded from
            this view
        filtered_fields (None): a set of field names of list fields that are
            filtered in this view
    """

    pass


class _PatchesView(fov.DatasetView):
    def __init__(
        self, source_collection, patches_stage, patches_dataset, _stages=None
    ):
        if _stages is None:
            _stages = []

        self._source_collection = source_collection
        self._patches_stage = patches_stage
        self._patches_dataset = patches_dataset
        self.__stages = _stages

    def __copy__(self):
        return self.__class__(
            self._source_collection,
            deepcopy(self._patches_stage),
            self._patches_dataset,
            _stages=deepcopy(self.__stages),
        )

    @property
    def _base_view(self):
        return self.__class__(
            self._source_collection,
            self._patches_stage,
            self._patches_dataset,
        )

    @property
    def _dataset(self):
        return self._patches_dataset

    @property
    def _root_dataset(self):
        return self._source_collection._root_dataset

    @property
    def _stages(self):
        return self.__stages

    @property
    def _all_stages(self):
        return (
            self._source_collection.view()._all_stages
            + [self._patches_stage]
            + self.__stages
        )

    @property
    def _id_field(self):
        if self._is_frames:
            return "frame_id"

        return "sample_id"

    @property
    def _label_fields(self):
        raise NotImplementedError("subclass must implement _label_fields")

    @property
    def name(self):
        return self.dataset_name + "-patches"

    @property
    def media_type(self):
        return fom.IMAGE

    def _tag_labels(self, tags, label_field, ids=None, label_ids=None):
        if label_field in self._label_fields:
            _ids = self.values("_" + self._id_field)

        _, label_ids = super()._tag_labels(
            tags, label_field, ids=ids, label_ids=label_ids
        )

        if label_field in self._label_fields:
            ids, label_ids = self._to_source_ids(label_field, _ids, label_ids)
            self._source_collection._tag_labels(
                tags, label_field, ids=ids, label_ids=label_ids
            )

    def _untag_labels(self, tags, label_field, ids=None, label_ids=None):
        if label_field in self._label_fields:
            _ids = self.values("_" + self._id_field)

        _, label_ids = super()._untag_labels(
            tags, label_field, ids=ids, label_ids=label_ids
        )

        if label_field in self._label_fields:
            ids, label_ids = self._to_source_ids(label_field, _ids, label_ids)
            self._source_collection._untag_labels(
                tags, label_field, ids=ids, label_ids=label_ids
            )

    def _to_source_ids(self, label_field, ids, label_ids):
        label_type = self._source_collection._get_label_field_type(label_field)
        is_list_field = issubclass(label_type, fol._LABEL_LIST_FIELDS)

        if not is_list_field:
            return ids, label_ids

        id_map = defaultdict(list)
        for _id, _label_id in zip(ids, label_ids):
            if etau.is_container(_label_id):
                id_map[_id].extend(_label_id)
            else:
                id_map[_id].append(_label_id)

        if not id_map:
            return [], []

        return zip(*id_map.items())

    def set_values(self, field_name, *args, **kwargs):
        field = field_name.split(".", 1)[0]
        must_sync = field in self._label_fields

        # The `set_values()` operation could change the contents of this view,
        # so we first record the sample IDs that need to be synced
        if must_sync and self._stages:
            ids = self.values("_id")
        else:
            ids = None

        super().set_values(field_name, *args, **kwargs)

        self._sync_source_field(field, ids=ids)

    def save(self, fields=None):
        """Saves the patches in this view to the underlying dataset.

        If this view contains any additional fields that were not extracted
        from the underlying dataset, these fields are not saved.

        This method **does not** delete patches from the underlying dataset
        that this view excludes.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.

        Args:
            fields (None): an optional field or list of fields to save. If
                specified, only these fields are overwritten
        """
        if etau.is_str(fields):
            fields = [fields]

        super().save(fields=fields)

        if fields is None:
            fields = self._label_fields
        else:
            fields = [l for l in fields if l in self._label_fields]

        self._sync_source_root(fields)

    def keep(self):
        """Deletes all patches that are **not** in this view from the
        underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """

        # The `keep()` operation below will delete patches, so we must sync
        # deletions to the source dataset first
        self._sync_source_root(self._label_fields, update=False, delete=True)

        super().keep()

    def keep_fields(self):
        """Deletes all patch field(s) that have been excluded from this view
        from the underlying dataset.

        .. note::

            This method is not a :class:`fiftyone.core.stages.ViewStage`;
            it immediately writes the requested changes to the underlying
            dataset.
        """
        self._sync_source_keep_fields()

        super().keep_fields()

    def reload(self):
        """Reloads this view from the source collection in the database.

        Note that :class:`PatchView` instances are not singletons, so any
        in-memory patches extracted from this view will not be updated by
        calling this method.
        """
        self._source_collection.reload()

        #
        # Regenerate the patches dataset
        #
        # This assumes that calling `load_view()` when the current patches
        # dataset has been deleted will cause a new one to be generated
        #

        self._patches_dataset.delete()
        _view = self._patches_stage.load_view(self._source_collection)
        self._patches_dataset = _view._patches_dataset

    def _sync_source_sample(self, sample):
        for field in self._label_fields:
            self._sync_source_sample_field(sample, field)

    def _sync_source_sample_field(self, sample, field):
        label_type = self._patches_dataset._get_label_field_type(field)
        is_list_field = issubclass(label_type, fol._LABEL_LIST_FIELDS)

        sample_id = sample[self._id_field]

        doc = sample._doc.field_to_mongo(field)
        if is_list_field:
            doc = doc[label_type._LABEL_LIST_FIELD]

        self._source_collection._set_labels(field, [sample_id], [doc])

    def _sync_source_field(self, field, ids=None):
        if field not in self._label_fields:
            return

        _, label_path = self._patches_dataset._get_label_field_path(field)

        if ids is not None:
            view = self._patches_dataset.mongo(
                [{"$match": {"_id": {"$in": ids}}}]
            )
        else:
            view = self._patches_dataset

        sample_ids, docs = view.aggregate(
            [foa.Values(self._id_field), foa.Values(label_path, _raw=True)]
        )

        self._source_collection._set_labels(field, sample_ids, docs)

    def _sync_source_root(self, fields, update=True, delete=False):
        for field in fields:
            self._sync_source_root_field(field, update=update, delete=delete)

    def _sync_source_root_field(self, field, update=True, delete=False):
        if field not in self._label_fields:
            return

        _, label_id_path = self._get_label_field_path(field, "id")
        label_path = label_id_path.rsplit(".", 1)[0]

        if update:
            sample_ids, docs = self._patches_dataset.aggregate(
                [
                    foa.Values(self._id_field),
                    foa.Values(label_path, _raw=True),
                ]
            )

            self._source_collection._set_labels(field, sample_ids, docs)

        if delete:
            all_ids = self._patches_dataset.values(label_id_path, unwind=True)
            self_ids = self.values(label_id_path, unwind=True)
            del_ids = set(all_ids) - set(self_ids)

            if del_ids:
                self._source_collection._delete_labels(
                    ids=del_ids, fields=field
                )

    def _sync_source_keep_fields(self):
        src_schema = self.get_field_schema()

        del_fields = set(self._label_fields) - set(src_schema.keys())
        if del_fields:
            self._source_collection.exclude_fields(del_fields).keep_fields()


class PatchesView(_PatchesView):
    """A :class:`fiftyone.core.view.DatasetView` of patches from a
    :class:`fiftyone.core.dataset.Dataset`.

    Patches views contain an ordered collection of patch samples, each of which
    contains a subset of a sample of the parent dataset corresponding to a
    single object or logical grouping of of objects.

    Patches retrieved from patches views are returned as :class:`PatchView`
    objects.

    Args:
        source_collection: the
            :class:`fiftyone.core.collections.SampleCollection` from which this
            view was created
        patches_stage: the :class:`fiftyone.core.stages.ToPatches` stage that
            defines how the patches were extracted
        patches_dataset: the :class:`fiftyone.core.dataset.Dataset` that serves
            the patches in this view
    """

    def __init__(
        self, source_collection, patches_stage, patches_dataset, _stages=None
    ):
        super().__init__(
            source_collection, patches_stage, patches_dataset, _stages=_stages
        )

        self._patches_field = patches_stage.field

    @property
    def _sample_cls(self):
        return PatchView

    @property
    def _label_fields(self):
        return [self._patches_field]

    @property
    def patches_field(self):
        """The field from which the patches in this view were extracted."""
        return self._patches_field


class EvaluationPatchesView(_PatchesView):
    """A :class:`fiftyone.core.view.DatasetView` containing evaluation patches
    from a :class:`fiftyone.core.dataset.Dataset`.

    Evalation patches views contain an ordered collection of evaluation
    examples, each of which contains the ground truth and/or predicted labels
    for a true positive, false positive, or false negative example from an
    evaluation run on the underlying dataset.

    Patches retrieved from patches views are returned as
    :class:`EvaluationPatchView` objects.

    Args:
        source_collection: the
            :class:`fiftyone.core.collections.SampleCollection` from which this
            view was created
        patches_stage: the :class:`fiftyone.core.stages.ToEvaluationPatches`
            stage that defines how the patches were extracted
        patches_dataset: the :class:`fiftyone.core.dataset.Dataset` that serves
            the patches in this view
    """

    def __init__(
        self, source_collection, patches_stage, patches_dataset, _stages=None
    ):
        super().__init__(
            source_collection, patches_stage, patches_dataset, _stages=_stages
        )

        eval_key = patches_stage.eval_key
        eval_info = source_collection.get_evaluation_info(eval_key)
        self._gt_field = eval_info.config.gt_field
        self._pred_field = eval_info.config.pred_field

    @property
    def _sample_cls(self):
        return EvaluationPatchView

    @property
    def _label_fields(self):
        return [self._gt_field, self._pred_field]

    @property
    def gt_field(self):
        """The ground truth field for the evaluation patches in this view."""
        return self._gt_field

    @property
    def pred_field(self):
        """The predictions field for the evaluation patches in this view."""
        return self._pred_field


def make_patches_dataset(
    sample_collection,
    field,
    other_fields=None,
    keep_label_lists=False,
    name=None,
):
    """Creates a dataset that contains one sample per object patch in the
    specified field of the collection.

    A ``sample_id`` field will be added that records the sample ID from which
    each patch was taken.

    By default, fields other than ``field`` and the default sample fields will
    not be included in the returned dataset.

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        field: the patches field, which must be of type
            :class:`fiftyone.core.labels.Detections` or
            :class:`fiftyone.core.labels.Polylines`
        other_fields (None): controls whether fields other than ``field`` and
            the default sample fields are included. Can be any of the
            following:

            -   a field or list of fields to include
            -   ``True`` to include all other fields
            -   ``None``/``False`` to include no other fields
        keep_label_lists (False): whether to store the patches in label list
            fields of the same type as the input collection rather than using
            their single label variants
        name (None): a name for the dataset

    Returns:
        a :class:`fiftyone.core.dataset.Dataset`
    """
    if sample_collection._is_frame_field(field):
        raise ValueError(
            "Frame label patches cannot be directly extracted; you must first "
            "convert your video dataset to frames via `to_frames()`"
        )

    fova.validate_collection(sample_collection, media_type=fom.IMAGE)

    if etau.is_str(other_fields):
        other_fields = [other_fields]

    is_frame_patches = sample_collection._is_frames
    patches_field = _get_patches_field(
        sample_collection, field, keep_label_lists
    )

    dataset = fod.Dataset(name=name, _patches=True, _frames=is_frame_patches)
    dataset.media_type = fom.IMAGE
    dataset.add_sample_field("sample_id", fof.ObjectIdField)
    dataset.create_index("sample_id")

    if is_frame_patches:
        dataset.add_sample_field("frame_id", fof.ObjectIdField)
        dataset.add_sample_field("frame_number", fof.FrameNumberField)
        dataset.create_index("frame_id")
        dataset.create_index([("sample_id", 1), ("frame_number", 1)])

    dataset.add_sample_field(field, **foo.get_field_kwargs(patches_field))

    if other_fields:
        src_schema = sample_collection.get_field_schema()
        curr_schema = dataset.get_field_schema()

        if other_fields == True:
            other_fields = [f for f in src_schema if f not in curr_schema]

        add_fields = [f for f in other_fields if f not in curr_schema]
        add_schema = {k: v for k, v in src_schema.items() if k in add_fields}
        dataset._sample_doc_cls.merge_field_schema(add_schema)

    _make_pretty_summary(dataset, is_frame_patches=is_frame_patches)

    patches_view = _make_patches_view(
        sample_collection,
        field,
        other_fields=other_fields,
        keep_label_lists=keep_label_lists,
    )
    _write_samples(dataset, patches_view)

    return dataset


def _get_patches_field(sample_collection, field_name, keep_label_lists):
    if keep_label_lists:
        return sample_collection.get_field(field_name)

    _, path = sample_collection._get_label_field_path(field_name)
    return sample_collection.get_field(path, leaf=True)


def make_evaluation_patches_dataset(
    sample_collection, eval_key, other_fields=None, name=None
):
    """Creates a dataset based on the results of the evaluation with the given
    key that contains one sample for each true positive, false positive, and
    false negative example in the input collection, respectively.

    True positive examples will result in samples with both their ground truth
    and predicted fields populated, while false positive/negative examples will
    only have one of their corresponding predicted/ground truth fields
    populated, respectively.

    If multiple predictions are matched to a ground truth object (e.g., if the
    evaluation protocol includes a crowd attribute), then all matched
    predictions will be stored in the single sample along with the ground truth
    object.

    The returned dataset will also have top-level ``type`` and ``iou`` fields
    populated based on the evaluation results for that example, as well as a
    ``sample_id`` field recording the sample ID of the example, and a ``crowd``
    field if the evaluation protocol defines a crowd attribute.

    .. note::

        The returned dataset will contain patches for the contents of the input
        collection, which may differ from the view on which the ``eval_key``
        evaluation was performed. This may exclude some labels that were
        evaluated and/or include labels that were not evaluated.

        If you would like to see patches for the exact view on which an
        evaluation was performed, first call
        :meth:`load_evaluation_view() <fiftyone.core.collections.SampleCollection.load_evaluation_view>`
        to load the view and then convert to patches.

    Args:
        sample_collection: a
            :class:`fiftyone.core.collections.SampleCollection`
        eval_key: an evaluation key that corresponds to the evaluation of
            ground truth/predicted fields that are of type
            :class:`fiftyone.core.labels.Detections` or
            :class:`fiftyone.core.labels.Polylines`
        other_fields (None): controls whether fields other than the
            ground truth/predicted fields and the default sample fields are
            included. Can be any of the following:

            -   a field or list of fields to include
            -   ``True`` to include all other fields
            -   ``None``/``False`` to include no other fields
        name (None): a name for the dataset

    Returns:
        a :class:`fiftyone.core.dataset.Dataset`
    """
    # Parse evaluation info
    eval_info = sample_collection.get_evaluation_info(eval_key)
    pred_field = eval_info.config.pred_field
    gt_field = eval_info.config.gt_field
    if hasattr(eval_info.config, "iscrowd"):
        crowd_attr = eval_info.config.iscrowd
    else:
        crowd_attr = None

    is_frame_patches = sample_collection._is_frames

    if is_frame_patches:
        if not pred_field.startswith(sample_collection._FRAMES_PREFIX):
            raise ValueError(
                "Cannot extract evaluation patches for sample-level "
                "evaluation '%s' from a frames view" % eval_key
            )

        pred_field = pred_field[len(sample_collection._FRAMES_PREFIX) :]
        gt_field = gt_field[len(sample_collection._FRAMES_PREFIX) :]
    elif sample_collection._is_frame_field(pred_field):
        raise ValueError(
            "Frame evaluation patches cannot be directly extracted; you must "
            "first convert your video dataset to frames via `to_frames()`"
        )

    if etau.is_str(other_fields):
        other_fields = [other_fields]

    _gt_field = sample_collection.get_field(gt_field)
    _pred_field = sample_collection.get_field(pred_field)

    # Setup dataset with correct schema
    dataset = fod.Dataset(name=name, _patches=True, _frames=is_frame_patches)
    dataset.media_type = fom.IMAGE
    dataset.add_sample_field("sample_id", fof.ObjectIdField)
    dataset.create_index("sample_id")

    if is_frame_patches:
        dataset.add_sample_field("frame_id", fof.ObjectIdField)
        dataset.add_sample_field("frame_number", fof.FrameNumberField)
        dataset.create_index("frame_id")
        dataset.create_index([("sample_id", 1), ("frame_number", 1)])

    dataset.add_sample_field(gt_field, **foo.get_field_kwargs(_gt_field))
    dataset.add_sample_field(pred_field, **foo.get_field_kwargs(_pred_field))

    if crowd_attr is not None:
        dataset.add_sample_field("crowd", fof.BooleanField)

    dataset.add_sample_field("type", fof.StringField)
    dataset.add_sample_field("iou", fof.FloatField)

    if other_fields:
        src_schema = sample_collection.get_field_schema()
        curr_schema = dataset.get_field_schema()

        if other_fields == True:
            other_fields = [f for f in src_schema if f not in curr_schema]

        add_fields = [f for f in other_fields if f not in curr_schema]
        add_schema = {k: v for k, v in src_schema.items() if k in add_fields}
        dataset._sample_doc_cls.merge_field_schema(add_schema)

    _make_pretty_summary(dataset, is_frame_patches=is_frame_patches)

    # Add ground truth patches
    gt_view = _make_eval_view(
        sample_collection,
        eval_key,
        gt_field,
        other_fields=other_fields,
        crowd_attr=crowd_attr,
    )
    _write_samples(dataset, gt_view)

    # Merge matched predictions
    _merge_matched_labels(dataset, sample_collection, eval_key, pred_field)

    # Add unmatched predictions
    unmatched_pred_view = _make_eval_view(
        sample_collection,
        eval_key,
        pred_field,
        other_fields=other_fields,
        skip_matched=True,
    )
    _add_samples(dataset, unmatched_pred_view)

    return dataset


def _make_pretty_summary(dataset, is_frame_patches=False):
    if is_frame_patches:
        set_fields = [
            "id",
            "sample_id",
            "frame_id",
            "filepath",
            "frame_number",
        ]
    else:
        set_fields = ["id", "sample_id", "filepath"]

    all_fields = dataset._sample_doc_cls._fields_ordered
    pretty_fields = set_fields + [f for f in all_fields if f not in set_fields]
    dataset._sample_doc_cls._fields_ordered = tuple(pretty_fields)


def _make_patches_view(
    sample_collection, field, other_fields=None, keep_label_lists=False
):
    label_type = sample_collection._get_label_field_type(field)
    if issubclass(label_type, _PATCHES_TYPES):
        list_field = field + "." + label_type._LABEL_LIST_FIELD
    else:
        raise ValueError(
            "Invalid label field type %s. Extracting patches is only "
            "supported for the following types: %s"
            % (label_type, _PATCHES_TYPES)
        )

    project = {
        "_id": False,
        "_media_type": True,
        "filepath": True,
        "metadata": True,
        "tags": True,
        field + "._cls": True,
        list_field: True,
    }

    if other_fields is not None:
        project.update({f: True for f in other_fields})

    if sample_collection._is_frames:
        project["_sample_id"] = True
        project["_frame_id"] = "$_id"
        project["frame_number"] = True
    else:
        project["_sample_id"] = "$_id"

    pipeline = [
        {"$project": project},
        {"$unwind": "$" + list_field},
        {"$set": {"_rand": {"$rand": {}}}},
        {"$set": {"_id": "$" + list_field + "._id"}},
    ]

    if keep_label_lists:
        pipeline.append({"$set": {list_field: ["$" + list_field]}})
    else:
        pipeline.append({"$set": {field: "$" + list_field}})

    return sample_collection.mongo(pipeline)


def _make_eval_view(
    sample_collection,
    eval_key,
    field,
    other_fields=None,
    skip_matched=False,
    crowd_attr=None,
):
    eval_type = field + "." + eval_key
    eval_id = field + "." + eval_key + "_id"
    eval_iou = field + "." + eval_key + "_iou"

    view = _make_patches_view(
        sample_collection, field, other_fields=other_fields
    )

    if skip_matched:
        view = view.mongo(
            [
                {
                    "$match": {
                        "$expr": {
                            "$or": [
                                {"$eq": ["$" + eval_id, _NO_MATCH_ID]},
                                {"$not": {"$gt": ["$" + eval_id, None]}},
                            ]
                        }
                    }
                }
            ]
        )

    view = view.mongo(
        [{"$set": {"type": "$" + eval_type, "iou": "$" + eval_iou}}]
    )

    if crowd_attr is not None:
        crowd_path1 = "$" + field + "." + crowd_attr

        # @todo can remove this when `Attributes` are deprecated
        crowd_path2 = "$" + field + ".attributes." + crowd_attr + ".value"

        view = view.mongo(
            [
                {
                    "$set": {
                        "crowd": {
                            "$cond": {
                                "if": {"$gt": [crowd_path1, None]},
                                "then": {"$toBool": crowd_path1},
                                "else": {
                                    "$cond": {
                                        "if": {"$gt": [crowd_path2, None]},
                                        "then": {"$toBool": crowd_path2},
                                        "else": None,
                                    }
                                },
                            }
                        }
                    }
                }
            ]
        )

    return _upgrade_labels(view, field)


def _upgrade_labels(view, field):
    tmp_field = "_" + field
    label_type = view._get_label_field_type(field)
    return view.mongo(
        [
            {"$set": {tmp_field: "$" + field}},
            {"$unset": field},
            {
                "$set": {
                    field: {
                        "_cls": label_type.__name__,
                        label_type._LABEL_LIST_FIELD: ["$" + tmp_field],
                    }
                }
            },
            {"$unset": tmp_field},
        ]
    )


def _merge_matched_labels(dataset, src_collection, eval_key, field):
    field_type = src_collection._get_label_field_type(field)

    list_field = field + "." + field_type._LABEL_LIST_FIELD
    eval_id = eval_key + "_id"
    eval_field = list_field + "." + eval_id

    src_collection._aggregate(
        post_pipeline=[
            {"$project": {list_field: True}},
            {"$unwind": "$" + list_field},
            {
                "$match": {
                    "$expr": {
                        "$and": [
                            {"$gt": ["$" + eval_field, None]},
                            {"$ne": ["$" + eval_field, _NO_MATCH_ID]},
                        ]
                    }
                }
            },
            {
                "$group": {
                    "_id": {"$toObjectId": "$" + eval_field},
                    "_labels": {"$push": "$" + list_field},
                }
            },
            {
                "$project": {
                    field: {
                        "_cls": field_type.__name__,
                        field_type._LABEL_LIST_FIELD: "$_labels",
                    }
                },
            },
            {
                "$merge": {
                    "into": dataset._sample_collection_name,
                    "on": "_id",
                    "whenMatched": "merge",
                    "whenNotMatched": "discard",
                }
            },
        ]
    )


def _write_samples(dataset, src_collection):
    src_collection._aggregate(
        detach_frames=True,
        detach_groups=True,
        post_pipeline=[{"$out": dataset._sample_collection_name}],
    )


def _add_samples(dataset, src_collection):
    src_collection._aggregate(
        detach_frames=True,
        detach_groups=True,
        post_pipeline=[
            {
                "$merge": {
                    "into": dataset._sample_collection_name,
                    "on": "_id",
                    "whenMatched": "keepExisting",
                    "whenNotMatched": "insert",
                }
            }
        ],
    )



"""
Utilities for working with datasets in
`OpenLABEL format <https://www.asam.net/index.php?eID=dumpFile&t=f&f=3876&token=413e8c85031ae64cc35cf42d0768627514868b2f>`_.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from collections import defaultdict
from copy import deepcopy
import enum
import logging
import os

import eta.core.serial as etas
import eta.core.utils as etau

import fiftyone.core.labels as fol
import fiftyone.core.metadata as fom
import fiftyone.core.utils as fou
import fiftyone.utils.data as foud


logger = logging.getLogger(__name__)


class SegmentationType(enum.Enum):
    """The FiftyOne label type used to store segmentations"""

    INSTANCE = 1
    POLYLINE = 2
    SEMANTIC = 3


class OpenLABELImageDatasetImporter(
    foud.LabeledImageDatasetImporter, foud.ImportPathsMixin
):
    """Importer for OpenLABEL image datasets stored on disk.

    See :ref:`this page <OpenLABELImageDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``data_path``
            and/or ``labels_path`` must be provided
        data_path (None): an optional parameter that enables explicit control
            over the location of the media. Can be any of the following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``dataset_dir`` where the media files reside
            -   an absolute directory path where the media files reside. In
                this case, the ``dataset_dir`` has no effect on the location of
                the data
            -   a filename like ``"data.json"`` specifying the filename of the
                JSON data manifest file in ``dataset_dir``
            -   an absolute filepath specifying the location of the JSON data
                manifest. In this case, ``dataset_dir`` has no effect on the
                location of the data
            -   a dict mapping file_ids to absolute filepaths

            If None, this parameter will default to whichever of ``data/`` or
            ``data.json`` exists in the dataset directory
        labels_path (None): an optional parameter that enables explicit control
            over the location of the labels. Can be any of the following:

            -   a filename like ``"labels.json"`` specifying the location of
                the labels in ``dataset_dir``
            -   a folder name like ``"labels"`` or ``"labels/"`` specifying a
                subfolder of ``dataset_dir`` where the multiple label files
                reside
            -   an absolute filepath to the labels. In this case,
                ``dataset_dir`` has no effect on the location of the labels

            If None, the parameter will default to looking for ``labels.json``
            and ``label/``
        label_types (None): a label type or list of label types to load. The
            supported values are
            ``("detections", "segmentations", "keypoints")``.
            By default, all labels are loaded
        use_polylines (False): whether to represent segmentations as
            :class:`fiftyone.core.labels.Polylines` instances rather than
            :class:`fiftyone.core.labels.Detections` with dense masks
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to load
        skeleton (None): a :class:`fiftyone.core.odm.dataset.KeypointSkeleton`
            to reference when loading keypoints
        skeleton_key (None): the key in the OpenLABEL annotations pointing to
            the label of each keypoint matching the labels defined in the given
            ``skeleton``
    """

    def __init__(
        self,
        dataset_dir=None,
        data_path=None,
        labels_path=None,
        label_types=None,
        use_polylines=False,
        shuffle=False,
        seed=None,
        max_samples=None,
        skeleton=None,
        skeleton_key=None,
    ):
        if dataset_dir is None and data_path is None and labels_path is None:
            raise ValueError(
                "At least one of `dataset_dir`, `data_path`, and "
                "`labels_path` must be provided"
            )

        data_path = self._parse_data_path(
            dataset_dir=dataset_dir,
            data_path=data_path,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=labels_path,
            default="labels.json",
        )

        _label_types = _parse_label_types(label_types)

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.data_path = data_path
        self.labels_path = labels_path
        self._label_types = _label_types
        self.use_polylines = use_polylines
        self.skeleton = skeleton
        self.skeleton_key = skeleton_key

        self._info = None
        self._image_paths_map = None
        self._annotations = None
        self._file_ids = None
        self._iter_file_ids = None

    def __iter__(self):
        self._iter_file_ids = iter(self._file_ids)
        return self

    def __len__(self):
        return len(self._file_ids)

    def __next__(self):
        file_id = next(self._iter_file_ids)

        if os.path.isfile(file_id):
            sample_path = file_id
        elif _remove_ext(file_id) in self._image_paths_map:
            sample_path = self._image_paths_map[_remove_ext(file_id)]
        else:
            sample_path = self._image_paths_map[
                _remove_ext(os.path.basename(file_id))
            ]

        seg_type = (
            SegmentationType.POLYLINE
            if self.use_polylines
            else SegmentationType.INSTANCE
        )

        width, height = self._annotations.get_dimensions(file_id)

        if height is None or width is None:
            sample_metadata = fom.ImageMetadata.build_for(sample_path)
            height, width = sample_metadata["height"], sample_metadata["width"]
        else:
            sample_metadata = fom.ImageMetadata(width=width, height=height)

        frame_size = (width, height)
        sample_labels, frame_labels = self._annotations.get_labels(
            file_id,
            self._label_types,
            frame_size,
            seg_type,
            skeleton=self.skeleton,
            skeleton_key=self.skeleton_key,
        )

        labels = _merge_frame_labels(sample_labels, frame_labels, seg_type)

        if self._has_scalar_labels:
            labels = next(iter(labels.values())) if labels else None

        return sample_path, sample_metadata, labels

    @property
    def has_dataset_info(self):
        return True

    @property
    def has_image_metadata(self):
        return True

    @property
    def _has_scalar_labels(self):
        return len(self._label_types) == 1

    @property
    def label_cls(self):
        return None

    def setup(self):
        image_paths_map = self._load_data_map(
            self.data_path, ignore_exts=True, recursive=True
        )

        file_ids = []
        annotations = OpenLABELAnnotations()

        if self.labels_path is not None:
            labels_path = fou.normpath(self.labels_path)

            base_dir = None
            if os.path.isfile(labels_path):
                label_paths = [labels_path]
            elif os.path.isdir(labels_path):
                base_dir = labels_path
            elif os.path.basename(
                labels_path
            ) == "labels.json" and os.path.isdir(_remove_ext(labels_path)):
                base_dir = _remove_ext(labels_path)
            else:
                label_paths = []

            if base_dir is not None:
                label_paths = etau.list_files(base_dir, recursive=True)
                label_paths = [l for l in label_paths if l.endswith(".json")]

            for label_path in label_paths:
                file_ids.extend(annotations.parse_labels(base_dir, label_path))

        file_ids = _validate_file_ids(file_ids, image_paths_map)

        self._info = {}
        self._image_paths_map = image_paths_map
        self._annotations = annotations
        self._file_ids = file_ids

    def get_dataset_info(self):
        return self._info


class OpenLABELVideoDatasetImporter(
    foud.LabeledVideoDatasetImporter, foud.ImportPathsMixin
):
    """Importer for OpenLABEL video datasets stored on disk.

    See :ref:`this page <OpenLABELVideoDataset-import>` for format details.

    Args:
        dataset_dir (None): the dataset directory. If omitted, ``data_path``
            and/or ``labels_path`` must be provided
        data_path (None): an optional parameter that enables explicit control
            over the location of the media. Can be any of the following:

            -   a folder name like ``"data"`` or ``"data/"`` specifying a
                subfolder of ``dataset_dir`` where the media files reside
            -   an absolute directory path where the media files reside. In
                this case, the ``dataset_dir`` has no effect on the location of
                the data
            -   a filename like ``"data.json"`` specifying the filename of the
                JSON data manifest file in ``dataset_dir``
            -   an absolute filepath specifying the location of the JSON data
                manifest. In this case, ``dataset_dir`` has no effect on the
                location of the data
            -   a dict mapping file_ids to absolute filepaths

            If None, this parameter will default to whichever of ``data/`` or
            ``data.json`` exists in the dataset directory
        labels_path (None): an optional parameter that enables explicit control
            over the location of the labels. Can be any of the following:

            -   a filename like ``"labels.json"`` specifying the location of
                the labels in ``dataset_dir``
            -   a folder name like ``"labels"`` or ``"labels/"`` specifying a
                subfolder of ``dataset_dir`` where the multiple label files
                reside
            -   an absolute filepath to the labels. In this case,
                ``dataset_dir`` has no effect on the location of the labels

            If None, the parameter will default to looking for ``labels.json``
            and ``labels/``
        label_types (None): a label type or list of label types to load. The
            supported values are
            ``("detections", "segmentations", "keypoints")``.
            By default, all labels are loaded
        use_polylines (False): whether to represent segmentations as
            :class:`fiftyone.core.labels.Polylines` instances rather than
            :class:`fiftyone.core.labels.Detections` with dense masks
        shuffle (False): whether to randomly shuffle the order in which the
            samples are imported
        seed (None): a random seed to use when shuffling
        max_samples (None): a maximum number of samples to load
        skeleton (None): a :class:`fiftyone.core.odm.dataset.KeypointSkeleton`
            to reference when loading keypoints
        skeleton_key (None): the key in the OpenLABEL annotations pointing to
            the label of each keypoint matching the labels defined in the given
            ``skeleton``
    """

    def __init__(
        self,
        dataset_dir=None,
        data_path=None,
        labels_path=None,
        label_types=None,
        use_polylines=False,
        shuffle=False,
        seed=None,
        max_samples=None,
        skeleton=None,
        skeleton_key=None,
    ):
        if dataset_dir is None and data_path is None and labels_path is None:
            raise ValueError(
                "At least one of `dataset_dir`, `data_path`, and "
                "`labels_path` must be provided"
            )

        data_path = self._parse_data_path(
            dataset_dir=dataset_dir,
            data_path=data_path,
            default="data/",
        )

        labels_path = self._parse_labels_path(
            dataset_dir=dataset_dir,
            labels_path=labels_path,
            default="labels.json",
        )

        _label_types = _parse_label_types(label_types)

        super().__init__(
            dataset_dir=dataset_dir,
            shuffle=shuffle,
            seed=seed,
            max_samples=max_samples,
        )

        self.data_path = data_path
        self.labels_path = labels_path
        self._label_types = _label_types
        self.use_polylines = use_polylines
        self.skeleton = skeleton
        self.skeleton_key = skeleton_key

        self._info = None
        self._video_paths_map = None
        self._annotations = None
        self._file_ids = None
        self._iter_file_ids = None

    def __iter__(self):
        self._iter_file_ids = iter(self._file_ids)
        return self

    def __len__(self):
        return len(self._file_ids)

    def __next__(self):
        file_id = next(self._iter_file_ids)

        if os.path.isfile(file_id):
            sample_path = file_id
        elif _remove_ext(file_id) in self._video_paths_map:
            sample_path = self._video_paths_map[_remove_ext(file_id)]
        else:
            sample_path = self._video_paths_map[
                _remove_ext(os.path.basename(file_id))
            ]

        width, height = self._annotations.get_dimensions(file_id)

        if height is None or width is None:
            sample_metadata = fom.VideoMetadata.build_for(sample_path)
            height, width = (
                sample_metadata["frame_height"],
                sample_metadata["frame_width"],
            )
        else:
            sample_metadata = fom.VideoMetadata(
                frame_width=width, frame_height=height
            )

        frame_size = (width, height)
        seg_type = (
            SegmentationType.POLYLINE
            if self.use_polylines
            else SegmentationType.INSTANCE
        )

        frame_size = (width, height)
        sample_labels, frame_labels = self._annotations.get_labels(
            file_id,
            self._label_types,
            frame_size,
            seg_type,
            skeleton=self.skeleton,
            skeleton_key=self.skeleton_key,
        )

        sample_labels = _remove_empty_labels(sample_labels, seg_type)

        return sample_path, sample_metadata, sample_labels, frame_labels

    @property
    def has_dataset_info(self):
        return True

    @property
    def has_video_metadata(self):
        return True

    @property
    def label_cls(self):
        return None

    def setup(self):
        video_paths_map = self._load_data_map(
            self.data_path, ignore_exts=True, recursive=True
        )

        file_ids = []
        annotations = OpenLABELAnnotations()

        if self.labels_path is not None:
            labels_path = fou.normpath(self.labels_path)

            base_dir = None
            if os.path.isfile(labels_path):
                label_paths = [labels_path]
            elif os.path.isdir(labels_path):
                base_dir = labels_path
            elif os.path.basename(
                labels_path
            ) == "labels.json" and os.path.isdir(_remove_ext(labels_path)):
                base_dir = _remove_ext(labels_path)
            else:
                label_paths = []

            if base_dir is not None:
                label_paths = etau.list_files(base_dir, recursive=True)
                label_paths = [l for l in label_paths if l.endswith(".json")]

            for label_path in label_paths:
                file_ids.extend(annotations.parse_labels(base_dir, label_path))

        file_ids = _validate_file_ids(file_ids, video_paths_map)

        self._info = {}
        self._video_paths_map = video_paths_map
        self._annotations = annotations
        self._file_ids = file_ids

    def get_dataset_info(self):
        return self._info


class OpenLABELAnnotations(object):
    """Annotations parsed from OpenLABEL format able to be converted to
    FiftyOne labels.
    """

    def __init__(self):
        self.objects = OpenLABELObjects()
        self.streams = OpenLABELStreams()
        self.metadata = {}

    def parse_labels(self, base_dir, labels_path):
        """Parses a single OpenLABEL labels file.

        Args:
            base_dir: path to the directory containing the labels file
            labels_path: path to the labels json file

        Returns:
            a list of potential file_ids that the parsed labels correspond to
        """
        abs_path = labels_path
        if not os.path.isabs(abs_path):
            abs_path = os.path.join(base_dir, labels_path)

        labels = etas.load_json(abs_path).get("openlabel", {})
        label_file_id = _remove_ext(labels_path)
        potential_file_ids = [label_file_id]

        metadata = OpenLABELMetadata(labels.get("metadata", {}))
        self.metadata[label_file_id] = metadata
        potential_file_ids.extend(metadata.parse_potential_file_ids())

        streams_dict = labels.get("streams", {})
        self.streams.parse_streams_dict(streams_dict, label_file_id)

        objects_dict = labels.get("objects", {})
        self.objects.parse_objects_dict(objects_dict, label_file_id)

        frames_dict = labels.get("frames", {})
        self._parse_frames(frames_dict, label_file_id)

        potential_file_ids.extend(self.streams.uris)

        return potential_file_ids

    def _parse_frames(self, frames, label_file_id):
        for frame_ind, frame in frames.items():
            frame_number = int(frame_ind) + 1

            objects = frame.get("objects", {})
            self.objects.parse_objects_dict(
                objects, label_file_id, frame_number=frame_number
            )

            streams_dict = frame.get("frame_properties", {}).get("streams", {})
            self.streams.parse_streams_dict(
                streams_dict, label_file_id, frame_number=frame_number
            )

    def get_dimensions(self, file_id):
        """Get the width and height of a given URI or file id

        Args:
            file_id: the unique identifier to a media file

        Returns:
            (width, height) of the given file
        """
        return self.streams.get_dimensions(file_id)

    def get_labels(
        self,
        uri,
        label_types,
        frame_size,
        seg_type,
        skeleton=None,
        skeleton_key=None,
    ):
        """Get the FiftyOne labels corresponding to the annotations of a given
        URI. The results are two dictionaries, sample- and frame-level, mapping
        field names to values and label objects.

        Args:
            uri: the unique identifier to a media file
            label_types: a list of label types to load. The
                supported values are
                ``("detections", "segmentations", "keypoints")``.
            frame_size: the (width, height) tuple for the media frame
            seg_type: the :class:`SegmentationType` to use for segmentation
                annotations
            skeleton (None): a
                :class:`fiftyone.core.odm.dataset.KeypointSkeleton` to use as a
                reference when loading keypoints
            skeleton_key (None): the name of the field in the OpenLABEL
                annotations containing the labels of keypoints matching the
                labels of the given skeleton

        Returns:
            a dictionary of sample level fields and label objects and a
            dictionary of frame numbers to frame level fields and label objects
        """
        stream_infos = self.streams.get_stream_info(uri)
        sample_objects = self.objects.get_objects(stream_infos)
        return sample_objects.to_labels(
            frame_size,
            label_types,
            seg_type,
            stream_infos,
            skeleton=skeleton,
            skeleton_key=skeleton_key,
        )


class OpenLABELStreamInfos(object):
    """A collection of multiple :class:`OpenLABELStreamInfo` objects."""

    def __init__(self, infos=None):
        self.infos = infos if infos else []

    def get_stream_attributes(self, frame_number=None):
        """Aggregates attributes from all streams in this collection.

        Args:
            frame_number (None): a specific frame number for which to get
                stream attributes

        Returns:
            a dictionary of attributes from all streams in this collection
        """
        attributes = {}
        for info in self.infos:
            is_sample = frame_number is None and info.is_sample_level
            has_frame_number = (
                info.frame_numbers and frame_number in info.frame_numbers
            )
            if is_sample or has_frame_number:
                attributes.update(info.get_stream_attributes())

        return attributes

    @property
    def frame_numbers(self):
        """All frame numbers existing in the :class:`OpenLABELStreamInfo`
        objects in this collection.
        """
        frame_numbers = []
        for info in self.infos:
            if info.frame_numbers:
                frame_numbers.extend(info.frame_numbers)
        return sorted(set(frame_numbers))


class OpenLABELStreamInfo(object):
    """Information about a stream used to gather specific objects for a media
    file.

    Args:
        frame_numbers (None): frame numbers related to this stream info
        stream (None): an :class:`OpenLABELStream`
        label_file_id (None): a label file id related to this stream info
        is_sample_level (None): whether this stream info corresponds to
            sample-level or frame-level
    """

    def __init__(
        self,
        frame_numbers=None,
        stream=None,
        label_file_id=None,
        is_sample_level=None,
    ):
        self.frame_numbers = frame_numbers
        self.stream = stream
        self.label_file_id = label_file_id
        self.is_sample_level = is_sample_level

    @property
    def is_streamless(self):
        """Whether there exists a stream corresponding to this info."""
        return self.stream is None

    def get_stream_attributes(self):
        """Get a dictionary of attributes for the stream in this object.

        Returns:
            a dictionary of attributes from the corresponding stream
        """
        attributes = {}
        if self.stream:
            attributes.update(self.stream.other_attrs)

        return attributes


class OpenLABELGroup(object):
    """A utility for parsing groups of OpenLABEL elements"""

    def __init__(self):
        self._element_id_to_element = {}
        self._keys_by_label_file_id = defaultdict(set)

    def _parse_group_dict(self, group_dict, label_file_id, frame_number=None):
        # Streams are unique by name and label file id
        for key, element_dict in group_dict.items():
            self._add_element_dict(
                label_file_id,
                key,
                element_dict,
                frame_number=frame_number,
            )

    @property
    def _element_type(self):
        raise NotImplementedError("Subclass must implement '_element_type'")

    @classmethod
    def _get_element_id(cls, label_file_id, name):
        return "%s_%s" % (label_file_id, name)

    @classmethod
    def _get_label_file_id(cls, element_id, element_name):
        return element_id[: -len(element_name) - 1]

    def _add_element_dict(self, label_file_id, key, info_d, frame_number=None):
        """Parses the given raw stream dictionary.

        Args:
            stream_name: the name of the stream being parsed
            stream_d: a dict containing stream information to parse
            frame_number (None): the frame number from which this stream
                information dict was parsed, 'None' if from the top-level
                streams
        """
        element_id = self._get_element_id(label_file_id, key)
        element = self._element_id_to_element.get(element_id, None)
        if element is None:
            element = self._element_type.from_anno_dict(
                key, info_d, frame_number=frame_number
            )
        else:
            element.update_dict(info_d, frame_number=frame_number)

        if element:
            self._element_id_to_element[element_id] = element
            self._keys_by_label_file_id[label_file_id].add(key)

        return element


class OpenLABELObjects(OpenLABELGroup):
    """A collection of :class:`OpenLABELObject` and corresponding utility
    methods.
    """

    @property
    def streams(self):
        """Get streams corresponding to any object in this collection."""
        _streams = []
        for obj in self.all_objects:
            _streams.extend(obj.streams)
        return list(set(_streams))

    @property
    def all_objects(self):
        """A list of :class:`OpenLABELObject` instances in this collection."""
        return list(self._element_id_to_element.values())

    def parse_objects_dict(
        self, objects_dict, label_file_id, frame_number=None
    ):
        """Parses the OpenLABEL annotations corresponding to a specific
        dictionary of objects.

        Args:
            objects_dict: the dict of OpenLABEL object annotations
            label_file_id: the name of the annotations file containing these
                objects
            frame_number (None): an optional frame that this ``objects_dict``
                is in
        """
        self._parse_group_dict(
            objects_dict, label_file_id, frame_number=frame_number
        )

    @property
    def _element_type(self):
        return OpenLABELObject

    def add_object(self, obj_key, label_file_id, obj):
        """Adds an :class:`OpenLABELObject` to this collection.

        Args:
            obj_key: the name of the object in the OpenLABEL annotations
            label_file_id: the filename of the annotations file containing this
                label
            obj: the :class:`OpenLABELObject` to add
        """
        obj_id = self._get_element_id(obj_key, label_file_id)
        self._element_id_to_element[obj_id] = obj

    def _get_filtered_object(self, obj_id, stream_info):
        # Get object that contains exactly the information needed, either with
        # frames filtered, removed, or merged
        if obj_id not in self._element_id_to_element:
            return None

        obj = self._element_id_to_element[obj_id]
        return obj.filter_stream(stream_info)

    def get_objects(self, stream_infos):
        """Gets any objects that correspond to an info in the given stream
        infos.

        Args:
            stream_infos: a :class:`OpenLABELStreamInfos` used to get
                corresponding objects

        Returns:
            an :class:`OpenLABELObjects` with objects that correspond to any of
            the given stream infos
        """
        stream_objects = OpenLABELObjects()
        for stream_info in stream_infos.infos:
            label_file_id = stream_info.label_file_id
            obj_keys = self._keys_by_label_file_id.get(label_file_id, [])
            for obj_key in obj_keys:
                obj_id = self._get_element_id(label_file_id, obj_key)
                obj = self._get_filtered_object(obj_id, stream_info)
                if obj:
                    stream_objects.add_object(obj_key, label_file_id, obj)

        return stream_objects

    def to_labels(
        self,
        frame_size,
        label_types,
        seg_type,
        stream_infos,
        skeleton=None,
        skeleton_key=None,
    ):
        """Converts the stored :class:`OpenLABELObject` to FiftyOne labels.

        Args:
            frame_size: the size of the image frame in pixels (width, height)
            label_types: a list of label types to load
            seg_type: the :class:`SegmentationType` to use to store
                segmentations
            stream_infos: the :class:`OpenLABELStreamInfos` containing
                sample-level attributes to parse into labels
            skeleton (None): a
                :class:`fiftyone.core.odm.dataset.KeypointSkeleton` to use when
                loading keypoint annotations
            skeleton_key (None): the name of the field in the OpenLABEL
                annotations containing the labels of keypoints matching the
                labels of the given skeleton

        Returns:
            a dictionary of sample level fields and label objects and a
            dictionary of frame numbers to frame level fields and label objects
        """
        frame_dets = defaultdict(list)
        frame_kps = defaultdict(list)
        frame_segs = defaultdict(list)
        for obj in self.all_objects:
            if "detections" in label_types:
                for frame_number, dets in obj.to_detections(
                    frame_size
                ).items():
                    frame_dets[frame_number].extend(dets)

            if "keypoints" in label_types:
                for frame_number, kps in obj.to_keypoints(
                    frame_size, skeleton=skeleton, skeleton_key=skeleton_key
                ).items():
                    frame_kps[frame_number].extend(kps)

            if "segmentations" in label_types:
                for frame_number, segs in obj.to_polylines(frame_size).items():
                    if seg_type == SegmentationType.POLYLINE:
                        _segs = segs
                    elif seg_type == SegmentationType.INSTANCE:
                        _segs = []
                        for seg in segs:
                            _segs.append(
                                seg.to_detection(frame_size=frame_size)
                            )
                    else:
                        raise NotImplementedError(
                            "Loading semantic segmentations is not yet "
                            "supported"
                        )
                    frame_segs[frame_number].extend(_segs)

        sample_labels = stream_infos.get_stream_attributes()

        frame_labels = defaultdict(dict)
        for frame_number in stream_infos.frame_numbers:
            frame_labels[frame_number] = stream_infos.get_stream_attributes(
                frame_number=frame_number
            )

        for frame_number, dets in frame_dets.items():
            frame_labels[frame_number]["detections"] = fol.Detections(
                detections=dets
            )

        for frame_number, kps in frame_kps.items():
            frame_labels[frame_number]["keypoints"] = fol.Keypoints(
                keypoints=kps
            )

        for frame_number, segs in frame_segs.items():
            if seg_type == SegmentationType.POLYLINE:
                frame_labels[frame_number]["segmentations"] = fol.Polylines(
                    polylines=segs
                )
            elif seg_type == SegmentationType.INSTANCE:
                frame_labels[frame_number]["segmentations"] = fol.Detections(
                    detections=segs
                )
            else:
                raise NotImplementedError(
                    "Loading semantic segmentations is not yet supported"
                )

        sample_labels.update(frame_labels.pop(None, {}))

        return sample_labels, dict(frame_labels)


class OpenLABELStreams(OpenLABELGroup):
    """A collection of OpenLABEL streams."""

    def __init__(self):
        super().__init__()
        self._uri_to_stream_ids = defaultdict(set)

    @property
    def uris(self):
        """All unique media file identifiers corresponding to streams in this
        collection.
        """
        _uris = []
        for stream in self._element_id_to_element.values():
            _uris.extend(stream.uris)
        return list(set(_uris))

    def parse_streams_dict(
        self, streams_dict, label_file_id, frame_number=None
    ):
        """Parses the OpenLABEL annotations corresponding to a specific
        dictionary of streams.

        Args:
            streams_dict: the dict of OpenLABEL stream annotations
            label_file_id: the name of the annotations file containing these
                streams
            frame_number (None): an optional frame that this ``streams_dict``
                is in
        """

        for key, element_dict in streams_dict.items():
            self._add_stream_dict(
                label_file_id,
                key,
                element_dict,
                frame_number=frame_number,
            )

    def get_dimensions(self, uri):
        """Get the width and height of a given URI or file id.

        Args:
            file_id: the unique identifier to a media file

        Returns:
            the ``(width, height)`` of the given file
        """
        stream_ids = list(self._uri_to_stream_ids.get(uri, []))
        # All streams pointing to this URI should be the same media
        if stream_ids:
            stream_id = stream_ids[0]
            stream = self._element_id_to_element.get(stream_id, None)
            if stream:
                return stream.width, stream.height

        return None, None

    @property
    def _element_type(self):
        return OpenLABELStream

    def _add_stream_dict(
        self, label_file_id, stream_name, stream_d, frame_number=None
    ):
        """Parses the given raw stream dictionary.

        Args:
            label_file_id: the filename of the annotations file from which
                these labels are parsed
            stream_name: the name of the stream being parsed
            stream_d: a dict containing stream information to parse
            frame_number (None): the frame number from which this stream
                information dict was parsed, 'None' if from the top-level
                streams
        """
        stream = self._add_element_dict(
            label_file_id, stream_name, stream_d, frame_number=frame_number
        )
        stream_id = self._get_element_id(label_file_id, stream_name)

        if stream is not None:
            for uri in stream.uris:
                self._uri_to_stream_ids[uri].add(stream_id)

    def get_stream_info(self, uri):
        """Get all stream infos, including stream and relevant frame numbers,
        for a given media file identifier.

        Args:
            uri: the unique media file identifer for which to get all stream
                infos

        Returns:
            the :class:`OpenLABELStreamInfos` corresponding to the given uri
        """
        infos = []
        if uri in self._uri_to_stream_ids:
            # Matches at least one stream at sample/frame level
            stream_ids = self._uri_to_stream_ids[uri]
            for stream_id in stream_ids:
                stream = self._element_id_to_element[stream_id]
                label_file_id = self._get_label_file_id(
                    stream_id,
                    stream.name,
                )
                frame_numbers, is_sample_level = stream.get_frame_numbers(uri)
                info = OpenLABELStreamInfo(
                    frame_numbers=frame_numbers,
                    stream=stream,
                    label_file_id=label_file_id,
                    is_sample_level=is_sample_level,
                )
                infos.append(info)
        else:
            # Does not match any stream, create info with only label file id,
            # must be sample level
            info = OpenLABELStreamInfo(label_file_id=uri, is_sample_level=True)
            infos.append(info)

        return OpenLABELStreamInfos(infos=infos)


class AttributeParser(object):
    """Methods used to parse attributes from OpenLABEL annotations."""

    _STREAM_KEYS = ["stream", "coordinate_system"]
    _IGNORE_KEYS = [
        "frame_intervals",
        "val",
        "attributes",
        "object_data",
        "object_data_pointers",
        "bbox",
        "point2d",
        "poly2d",
    ]

    @classmethod
    def _parse_attributes(cls, d):
        attributes = {k: v for k, v in d.items() if k not in cls._IGNORE_KEYS}
        attributes_dict = d.get("attributes", {})
        stream = None
        for k in cls._STREAM_KEYS:
            if k in d:
                stream = d[k]

        _attrs, _stream = cls._parse_name_val_attributes(attributes_dict)
        attributes.update(_attrs)
        if _stream:
            stream = _stream

        return attributes, stream

    @classmethod
    def _parse_name_val_attributes(cls, d):
        attributes = {}
        stream = None
        for attr_type, attrs in d.items():
            for attr in attrs:
                name = attr["name"]
                val = attr["val"]
                if name.lower() in cls._STREAM_KEYS:
                    stream = val

                if name.lower() not in cls._IGNORE_KEYS:
                    attributes[name] = val
        return attributes, stream


class OpenLABELShape(AttributeParser):
    """A single OpenLABEL shape like a bounding box or polygon."""

    def __init__(self, coords, attributes=None, stream=None):
        self.coords = coords
        self.attributes = attributes if attributes else {}
        self.stream = stream

    @classmethod
    def from_shape_dict(cls, d):
        """Constructs a shape from a dictionary of information.

        Args:
            d: a dictionary containing information about a shape
        Returns:
            an :class:`OpenLABELShape`
        """
        coords = d.pop("val", None)
        attributes, stream = cls._parse_attributes(d)
        return cls(coords, attributes=attributes, stream=stream)


class OpenLABELBBox(OpenLABELShape):
    """An OpenLABEL bounding box."""

    def to_label(self, label, attributes, width, height):
        """Convert this shape to a FiftyOne label.

        Args:
            label: the class label for this shape
            attributes: a dictionary of attributes for this shape
            width: the width of the frame in pixels
            height: the height of the frame in pixels

        Return:
            an :class:`fiftyone.core.labels.Detection`
        """
        num_coords = len(self.coords)
        if num_coords != 4:
            raise ValueError(
                "Expected bounding box to have 4 coordinates, found %d"
                % num_coords
            )

        cx, cy, w, h = self.coords
        x = cx - (w / 2)
        y = cy - (h / 2)
        bounding_box = [x / width, y / height, w / width, h / height]

        _attrs = deepcopy(attributes)
        _attrs.update(self.attributes)

        return fol.Detection(
            label=label,
            bounding_box=bounding_box,
            **_attrs,
        )


class OpenLABELPoly2D(OpenLABELShape):
    """An OpenLABEL polyline."""

    def to_label(self, label, attributes, width, height):
        """Convert this shape to a FiftyOne label.

        Args:
            label: the class label for this shape
            attributes: a dictionary of attributes for this shape
            width: the width of the frame in pixels
            height: the height of the frame in pixels

        Return:
            an :class:`fiftyone.core.labels.Polyline`
        """
        rel_points = [
            [(x / width, y / height) for x, y, in _pairwise(self.coords)]
        ]
        _attrs = deepcopy(attributes)
        _attrs.update(self.attributes)

        filled = _attrs.pop("filled", None)
        if filled is None:
            filled = not _attrs.get("is_hole", True)

        closed = _attrs.pop("closed", True)
        _attrs.pop("label", None)

        return fol.Polyline(
            label=label,
            points=rel_points,
            filled=filled,
            closed=closed,
            **_attrs,
        )


class OpenLABELPoint(OpenLABELShape):
    """An OpenLABEL keypoint."""

    @classmethod
    def _sort_by_skeleton(cls, points, attrs, label_order, skeleton_order):
        if len(points) != len(label_order):
            return points, attrs

        if not isinstance(skeleton_order, list):
            skeleton_order = skeleton_order.labels

        sorted_points = []
        sorted_attrs = {}
        attrs_to_sort = {}
        for k, v in attrs.items():
            if isinstance(v, list) and len(v) == len(points):
                attrs_to_sort[k] = v
                sorted_attrs[k] = []
            else:
                sorted_attrs[k] = v

        for label in skeleton_order:
            if label not in label_order:
                sorted_points.append([float("nan"), float("nan")])
                for k in attrs_to_sort:
                    sorted_attrs[k].append(None)
            else:
                ind = label_order.index(label)
                sorted_points.append(points[ind])
                for k, v in attrs_to_sort.items():
                    sorted_attrs[k].append(v[ind])

        return sorted_points, sorted_attrs

    def to_label(
        self,
        label,
        attributes,
        width,
        height,
        skeleton=None,
        skeleton_key=None,
    ):
        """Convert this shape to a FiftyOne label.

        Args:
            label: the class label for this shape
            attributes: a dictionary of attributes for this shape
            width: the width of the frame in pixels
            height: the height of the frame in pixels
            skeleton (None): a
                :class:`fiftyone.core.odm.dataset.KeypointSkeleton` used to
                sort list attributes based on the labels in the skeleton. Used
                only if ``skeleton_key`` is provided
            skeleton_key (None): the string key into the attributes dictionary
                containing the label of each point, used to sort list attribute
                fields based on the labels in the skeleton. Used only if
                ``skeleton`` is provided

        Return:
            an :class:`fiftyone.core.labels.Keypoint`
        """
        rel_points = [(x / width, y / height) for x, y, in self.coords]
        _attrs = deepcopy(attributes)
        _attrs.update(self.attributes)
        if skeleton and skeleton_key and skeleton_key in _attrs:
            label_order = _attrs.pop(skeleton_key)
            rel_points, _attrs = self._sort_by_skeleton(
                rel_points, _attrs, label_order, skeleton
            )

        return fol.Keypoint(label=label, points=rel_points, **_attrs)


class OpenLABELShapes(AttributeParser):
    """A collection of OpenLABEL shapes."""

    def __init__(self, shapes=None, attributes=None, stream=None):
        self.shapes = shapes if shapes else []
        self.attributes = attributes if attributes else {}
        self.stream = stream

    @property
    def streams(self):
        """A list of streams corresponding to any object in this collection."""
        streams = []
        if self.stream:
            streams.append(self.stream)

        for shape in self.shapes:
            stream = shape.stream
            if stream:
                streams.append(stream)

        return streams

    @classmethod
    def from_object_data_list(cls, shape_type, l, attributes=None):
        """Construct an :class:`OpenLABELShapes` from a list of shape
        dictionaries.

        Args:
            shape_type: the type of the shape being loaded. Options are
                (:class:`OpenLABELBBox`, :class:`OpenLABELPoly2D`,
                :class:`OpenLABELPoint`)
            l: a list of shape dictionaries parsed from OpenLABEL object
                annotations
            attributes (None): a dictionary of attributes corresponding to all
                shapes in this collection

        Returns:
            a :class:`OpenLABELShapes`
        """
        shapes = []
        for shape_d in l:
            shapes.append(shape_type.from_shape_dict(shape_d))

        stream = None
        if attributes:
            attributes, stream = cls._parse_name_val_attributes(attributes)

        return cls(shapes=shapes, attributes=attributes, stream=stream)

    def merge_shapes(self, shapes):
        """Merges another :class:`OpenLABELShapes` into this one.

        Args:
            shapes: another :class:`OpenLABELShapes` to merge into this
                object
        """
        if shapes:
            self.shapes.extend(shapes.shapes)
            self.attributes.update(shapes.attributes)

    def to_labels(
        self,
        label,
        attributes,
        width,
        height,
        is_points=False,
        skeleton=None,
        skeleton_key=None,
    ):
        """Convert this shape to a FiftyOne label.

        Args:
            label: the class label for this shape
            attributes: a dictionary of attributes for this shape
            width: the width of the frame in pixels
            height: the height of the frame in pixels
            is_points (False): whether the labels being converted are keypoints
            skeleton (None): a
                :class:`fiftyone.core.odm.dataset.KeypointSkeleton` used to
                sort list attributes based on the labels in the skeleton. Used
                only if ``is_points`` and ``skeleton_key`` is provided
            skeleton_key (None): the string key into the attributes dictionary
                containing the label of each point, used to sort list attribute
                fields based on the labels in the skeleton. Used only if
                ``is_points`` and ``skeleton`` is provided

        Return:
            an :class:`fiftyone.core.labels.Keypoint`
        """
        if is_points:
            return self._to_point_labels(
                label,
                attributes,
                width,
                height,
                skeleton=skeleton,
                skeleton_key=skeleton_key,
            )

        return self._to_individual_labels(label, attributes, width, height)

    @property
    def _homogenous_shape_types(self):
        types = [type(s) for s in self.shapes]

        if len(set(types)) > 1:
            return False

        return True

    def _to_point_labels(
        self,
        label,
        attributes,
        width,
        height,
        skeleton=None,
        skeleton_key=None,
    ):
        labels = []

        if not self.shapes:
            return labels

        if not self._homogenous_shape_types or not isinstance(
            self.shapes[0], OpenLABELPoint
        ):
            raise ValueError(
                "Found non-point shapes when attempting to convert to "
                "Keypoint labels."
            )

        # Convert keypoints to list of points
        coords = []
        _attrs = defaultdict(list)
        stream = None
        for shape in self.shapes:
            coords.append(shape.coords)
            for k, v in shape.attributes.items():
                _attrs[k].append(v)

            if shape.stream:
                stream = shape.stream

        if coords:
            _attrs = dict(_attrs)
            _attrs.update(self.attributes)
            shape = type(self.shapes[0])(
                coords, attributes=dict(_attrs), stream=stream
            )
            labels.append(
                shape.to_label(
                    label,
                    attributes,
                    width,
                    height,
                    skeleton=skeleton,
                    skeleton_key=skeleton_key,
                )
            )

        return labels

    def _to_individual_labels(self, label, attributes, width, height):
        labels = []
        _attrs = deepcopy(attributes)
        _attrs.update(self.attributes)
        for shape in self.shapes:
            labels.append(shape.to_label(label, _attrs, width, height))

        return labels


class OpenLABELStream(object):
    """An OpenLABEL stream corresponding to one uri or file_id.

    Args:
        name (None): the name of the stream
        type (None): the type of the stream
        description (None): a string description for this stream
        properties (None): a dict of properties for this stream
        uri (None): the uri or file_id of the media corresponding to this
            stream
        other_attrs (None): a dictionary of other attributes corresponding to
            this stream
    """

    _HEIGHT_KEYS = ["height", "height_px"]
    _WIDTH_KEYS = ["width", "width_px"]
    _URI_KEYS = ["uri"]

    def __init__(
        self,
        name,
        type=None,
        properties=None,
        uris=None,
        other_attrs=None,
    ):
        self.name = name
        self.type = type
        self.properties = properties
        self.height = None
        self.width = None
        self.other_attrs = other_attrs if other_attrs else {}

        self._uris = uris if uris else []

        self.frame_streams = {}

        if properties:
            self._parse_properties_dict(properties)

    def _parse_properties_dict(self, d):
        for k, v in d.items():
            if etau.is_numeric(v):
                self._check_height_width(k, v)
            elif isinstance(v, dict):
                self._parse_properties_dict(v)

    def _check_height_width(self, key, value):
        if key.lower() in self._HEIGHT_KEYS:
            self.height = float(value)

        if key.lower() in self._WIDTH_KEYS:
            self.width = float(value)

    def update_dict(self, d, frame_number=None):
        """Updates this stream with additional information.

        Args:
            d: a dict containing additional stream information
            frame_number (None): the frame number from which this stream
                information dict was parsed, 'None' if from the top-level
                streams
        """
        if frame_number:
            frame_stream = self.frame_streams.get(
                frame_number, OpenLABELStream(self.name)
            )
            frame_stream.update_dict(d)
            self.frame_streams[frame_number] = frame_stream

        else:
            _type, properties, uris, other_attrs = self._parse_stream_dict(d)

            if _type:
                if _type != "camera":
                    return

                self.type = _type

            if properties:
                self.properties = properties
                self._parse_properties_dict(properties)

            if uris:
                self._uris = sorted(set(self._uris + uris))

            if other_attrs:
                self.other_attrs.update(other_attrs)

    @property
    def uris(self):
        """Get uris corresponding to any stream in this collection."""
        _uris = deepcopy(self._uris)
        for _stream in self.frame_streams.values():
            _uris.extend(_stream.uris)

        return sorted(set(_uris))

    def get_frame_numbers(self, uri):
        """Get frame numbers corresponding to the given uri."""
        is_sample_level = False
        if uri in self._uris:
            # URI corresponds to all frames in this stream
            # Likely a video annotation
            # For an image, that sample will just contain annotations from all
            # frames
            is_sample_level = True
            return list(self.frame_streams.keys()), is_sample_level

        frame_numbers = []
        for frame_number, frame_stream in self.frame_streams.items():
            # URI exists in one or more frames in this stream
            # Likely an image annotation
            if uri in frame_stream.uris:
                frame_numbers.append(frame_number)

        return frame_numbers, is_sample_level

    @classmethod
    def from_anno_dict(cls, stream_name, d, frame_number):
        """Create an OpenLABEL stream from the stream information dictionary.

        Args:
            stream_name: the name of the stream
            d: a dict containing information about this stream
            frame_number: the frame number from which this stream
                information dict was parsed, 'None' if from the top-level
                streams

        Returns:
            An :class:`OpenLABELStream`
        """
        if frame_number is not None:
            stream = cls(stream_name)
            stream.update_dict(d, frame_number=frame_number)
        else:
            _type, properties, uris, other_attrs = cls._parse_stream_dict(d)
            if _type and _type != "camera":
                return None

            stream = cls(
                stream_name,
                type=_type,
                properties=properties,
                uris=uris,
                other_attrs=other_attrs,
            )

        return stream

    @classmethod
    def _parse_stream_dict(cls, d):
        _type = d.pop("type", None)
        properties = d.pop("stream_properties", None)

        uris = []
        for uri_key in cls._URI_KEYS:
            uri_val = d.pop(uri_key, None)
            if uri_val and uri_val not in uris:
                uris.append(uri_val)

        return _type, properties, uris, d


class OpenLABELMetadata(object):
    """A parser and storage for OpenLABEL metadata."""

    _POTENTIAL_FILENAME_KEYS = ["file_id", "uri", "file_id", "filepath"]

    def __init__(self, metadata_dict):
        self.metadata_dict = metadata_dict
        self._parse_seg_type()

    def _parse_seg_type(self):
        # Currently unused
        self.seg_type = SegmentationType.INSTANCE
        if "annotation_type" in self.metadata_dict:
            if (
                self.metadata_dict["annotation_type"]
                == "semantic segmentation"
            ):
                self.seg_type = SegmentationType.SEMANTIC

    def parse_potential_file_ids(self):
        """Parses metadata for any fields that may correspond to a label-wide
        media file_id.

        Returns:
            a list of potential file_id strings
        """
        file_ids = []
        for k, v in self.metadata_dict.items():
            if k.lower() in self._POTENTIAL_FILENAME_KEYS:
                file_ids.append(v)

        return file_ids


class OpenLABELObject(AttributeParser):
    """An object parsed from OpenLABEL labels.

    Args:
        key: the OpenLABEL key string for this object
        name (None): the name string of the object
        type (None): the type string of the object
        bboxes (None): an :class`OpenLABELShapes` of bounding boxes for this
            object
        segmentations (None): an :class`OpenLABELShapes` of polygon
            segmentations for this object
        keyponts (None): an :class:`OpenLABELShapes` of keypoints for this
            object
        stream (None): the :class:`OpenLABELStream` this object corresponds to
        other_attrs (None): a dict of attributes and their values for this
            object
        is_frame_level (False): whether this object is sample-level or
            frame-level
    """

    _STREAM_KEYS = ["stream", "coordinate_system"]

    def __init__(
        self,
        key,
        name=None,
        type=None,
        bboxes=None,
        segmentations=None,
        keypoints=None,
        stream=None,
        other_attrs=None,
        is_frame_level=False,
    ):

        self.shapes = {
            "bboxes": OpenLABELShapes(),
            "segmentations": OpenLABELShapes(),
            "keypoints": OpenLABELShapes(),
        }
        if bboxes:
            self.shapes["bboxes"] = bboxes

        if segmentations:
            self.shapes["segmentations"] = segmentations

        if keypoints:
            self.shapes["keypoints"] = keypoints

        self.key = key
        self.name = name
        self.type = type

        self.stream = stream
        self.other_attrs = other_attrs if other_attrs else {}
        self.frame_objects = {}
        self.is_frame_level = is_frame_level

    @property
    def _sample_level_streams(self):
        _streams = []
        if self.stream:
            _streams.append(self.stream)

        for _shapes in self.shapes.values():
            _streams.extend(_shapes.streams)

        return list(set(_streams))

    @property
    def streams(self):
        """Get streams corresponding to this object."""
        _streams = deepcopy(self._sample_level_streams)

        for _object in self.frame_objects.values():
            _streams.extend(_object.streams)

        return list(set(_streams))

    @property
    def is_streamless(self):
        """Whether any streams are connected to this object or corresponding
        frame-level objects.
        """
        return not bool(self._sample_level_streams)

    def filter_stream(self, stream_info):
        """Filters this object to contain only frame labels specified in the
        given stream info

        Args:
            stream_info: the :class:`OpenLABELStreamInfo` to use to filter this
                object

        Returns:
            an :class:`OpenLABELObject` containing only frames related to the
            given stream info
        """
        # if obj is top level and stream info is sample level
        if stream_info.is_streamless:
            if self.is_frame_level or not self.is_streamless:
                return None

            # This object is sample-level and streamless
            return self

        if stream_info.is_sample_level:
            return self

        return self._keep_frames(stream_info.frame_numbers)

    def _keep_frames(self, frame_numbers):
        _obj = deepcopy(self)
        numbers_to_remove = set(_obj.frame_objects.keys()) - set(frame_numbers)
        for frame_number in numbers_to_remove:
            _obj.frame_objects.pop(frame_number, None)
        return _obj

    def _to_labels(
        self,
        frame_size,
        shape_type,
        parent=None,
        is_points=False,
        skeleton=None,
        skeleton_key=None,
    ):
        label, attributes, width, height = self._get_label_attrs(
            frame_size, parent=parent
        )

        frame_labels = defaultdict(list)
        shapes = self.shapes[shape_type]
        frame_labels[None] = shapes.to_labels(
            label,
            attributes,
            width,
            height,
            is_points=is_points,
            skeleton=skeleton,
            skeleton_key=skeleton_key,
        )

        for frame_number, frame_object in self.frame_objects.items():
            frame_labels[frame_number].extend(
                frame_object._to_labels(
                    frame_size,
                    shape_type,
                    parent=self,
                    is_points=is_points,
                    skeleton=skeleton,
                    skeleton_key=skeleton_key,
                )[None]
            )

        return frame_labels

    def to_detections(self, frame_size):
        """Converts the bounding boxes in this object to
        :class:`fiftyone.core.labels.Detection` objects.

        Args:
            frame_size: the size of the frame in pixels (width, height)

        Returns:
            a list of :class:`fiftyone.core.labels.Detection` objects for each
            bounding box in this object
        """
        return self._to_labels(frame_size, "bboxes")

    def _get_label_attrs(self, frame_size, parent=None):
        label = self.type
        if label is None and parent:
            label = parent.type

        attributes = self._get_object_attributes(parent=parent)

        width, height = frame_size
        return label, attributes, width, height

    def to_polylines(self, frame_size):
        """Converts the segmentations in this object to
        :class:`fiftyone.core.labels.Polyline` objects.

        Args:
            frame_size: the size of the frame in pixels (width, height)

        Returns:
            a list of :class:`fiftyone.core.labels.Polyline` objects for each
            polyline in this object
        """
        return self._to_labels(frame_size, "segmentations")

    def to_keypoints(self, frame_size, skeleton=None, skeleton_key=None):
        """Converts the keypoints in this object to
        :class:`fiftyone.core.labels.Keypoint` objects.

        Args:
            frame_size: the size of the frame in pixels (width, height)

        Returns:
            a list of :class:`fiftyone.core.labels.Keypoint` objects for each
            keypoint in this object
        """
        return self._to_labels(
            frame_size,
            "keypoints",
            is_points=True,
            skeleton=skeleton,
            skeleton_key=skeleton_key,
        )

    @classmethod
    def from_anno_dict(cls, obj_key, d, frame_number=None):
        """Create an :class:`OpenLABELObject` from the raw label dictionary.

        Args:
            anno_id: id of the object
            d: dict containing the information for this object

        Returns:
            a tuple containing the :class:`OpenLABELObject` and the frame
            numbers the object corresponds to, if any.
        """
        if frame_number is not None:
            obj = cls(obj_key, is_frame_level=False)
            obj.update_dict(d, frame_number=frame_number)
        else:
            (
                bboxes,
                segmentations,
                points,
                name,
                _type,
                stream,
                other_attrs,
            ) = cls._parse_object_dict(d)

            obj = cls(
                obj_key,
                name=name,
                type=_type,
                bboxes=bboxes,
                segmentations=segmentations,
                keypoints=points,
                stream=stream,
                other_attrs=other_attrs,
                is_frame_level=False,
            )
        return obj

    @classmethod
    def _get_shape_list(cls, object_data, key):
        l = object_data.pop(key, [])
        if isinstance(l, dict):
            l = [l]
        return l

    @classmethod
    def _parse_object_dict(cls, d):
        object_data = d.pop("object_data", {})

        bbox_l = cls._get_shape_list(object_data, "bbox")
        poly2d_l = cls._get_shape_list(object_data, "poly2d")
        point2d_l = cls._get_shape_list(object_data, "point2d")

        name = d.pop("name", None)
        _type = d.pop("type", None)
        attributes, stream = cls._parse_attributes(d)

        bboxes = OpenLABELShapes.from_object_data_list(
            OpenLABELBBox, bbox_l, attributes=object_data
        )
        segmentations = OpenLABELShapes.from_object_data_list(
            OpenLABELPoly2D, poly2d_l, attributes=object_data
        )
        points = OpenLABELShapes.from_object_data_list(
            OpenLABELPoint, point2d_l, attributes=object_data
        )

        return (
            bboxes,
            segmentations,
            points,
            name,
            _type,
            stream,
            attributes,
        )

    def update_dict(self, d, frame_number=None):
        """Updates this :class:`OpenLABELObject` given the raw label
        dictionary.

        Args:
            d: dict containing the information for this object

        Returns:
            newly parsed frame numbers the object corresponds to, if any
        """
        if frame_number:
            frame_object = self.frame_objects.get(
                frame_number, OpenLABELObject(self.key, is_frame_level=True)
            )
            frame_object.update_dict(d)
            self.frame_objects[frame_number] = frame_object
        else:
            (
                bboxes,
                segmentations,
                points,
                name,
                _type,
                stream,
                other_attrs,
            ) = self._parse_object_dict(d)

            self.shapes["bboxes"].merge_shapes(bboxes)
            self.shapes["segmentations"].merge_shapes(segmentations)
            self.shapes["keypoints"].merge_shapes(points)

            if name and not self.name:
                self.name = name

            if stream and not self.stream:
                self.stream = stream

            self.other_attrs.update(other_attrs)

    def _get_object_attributes(self, parent=None):
        attributes = {}

        if parent:
            attributes.update(parent._get_object_attributes())

        if self.name is not None:
            attributes["name"] = self.name

        if self.key is not None:
            attributes["OpenLABEL_id"] = self.key

        attributes.update(self.other_attrs)

        return attributes


def _validate_file_ids(potential_file_ids, sample_paths_map):
    file_ids = []
    potential_file_ids = set(potential_file_ids)
    if None in potential_file_ids:
        potential_file_ids.remove(None)

    for file_id in potential_file_ids:
        is_file = os.path.exists(file_id)
        has_file_id = _remove_ext(file_id) in sample_paths_map
        has_basename = (
            _remove_ext(os.path.basename(file_id)) in sample_paths_map
        )
        if is_file or has_file_id or has_basename:
            file_ids.append(file_id)

    return file_ids


def _parse_label_types(label_types):
    if label_types is None:
        return _SUPPORTED_LABEL_TYPES

    if etau.is_str(label_types):
        label_types = [label_types]
    else:
        label_types = list(label_types)

    bad_types = [l for l in label_types if l not in _SUPPORTED_LABEL_TYPES]

    if len(bad_types) == 1:
        raise ValueError(
            "Unsupported label type '%s'. Supported types are %s"
            % (bad_types[0], _SUPPORTED_LABEL_TYPES)
        )

    if len(bad_types) > 1:
        raise ValueError(
            "Unsupported label types %s. Supported types are %s"
            % (bad_types, _SUPPORTED_LABEL_TYPES)
        )

    return label_types


_SUPPORTED_LABEL_TYPES = [
    "detections",
    "segmentations",
    "keypoints",
]


def _pairwise(x):
    y = iter(x)
    return zip(y, y)


def _remove_ext(p):
    return os.path.splitext(p)[0]


def _merge_frame_labels(sample_labels, frame_labels, seg_type):
    # Add frame labels to sample labels, if there is a key collision, merge the
    # labels if they are a list field otherewise skip
    for labels in frame_labels.values():
        for name, value in labels.items():
            if name == "detections":
                _extend_labels(
                    sample_labels, value, name, fol.Detections, "detections"
                )

            elif name == "keypoints":
                _extend_labels(
                    sample_labels, value, name, fol.Keypoints, "keypoints"
                )

            elif name == "segmentations":
                if seg_type == SegmentationType.POLYLINE:
                    _extend_labels(
                        sample_labels, value, name, fol.Polylines, "polylines"
                    )
                elif seg_type == SegmentationType.INSTANCE:
                    _extend_labels(
                        sample_labels,
                        value,
                        name,
                        fol.Detections,
                        "detections",
                    )
                else:
                    raise NotImplementedError(
                        "Loading semantic segmentations is not yet supported"
                    )

            elif name in sample_labels:
                if isinstance(sample_labels[name], list):
                    if not isinstance(value, list):
                        value = [value]

                    sample_labels[name].extend(value)

            else:
                sample_labels[name] = value

    return sample_labels


def _extend_labels(sample_labels, value, name, labels_type, label_kwarg):
    if name not in sample_labels:
        sample_labels[name] = labels_type(**{label_kwarg: []})
    else:
        sample_labels[name][label_kwarg].extend(value[label_kwarg])


def _remove_empty_labels(sample_labels, seg_type):
    _remove_empty_label_type(sample_labels, "detections", "detections")
    _remove_empty_label_type(sample_labels, "keypoints", "keypoints")
    if seg_type == SegmentationType.POLYLINE:
        _remove_empty_label_type(sample_labels, "segmentations", "polylines")
    elif seg_type == SegmentationType.INSTANCE:
        _remove_empty_label_type(sample_labels, "segmentations", "detections")
    else:
        raise NotImplementedError(
            "Loading semantic segmentations is not yet supported"
        )

    return sample_labels


def _remove_empty_label_type(sample_labels, key, label_type):
    if key in sample_labels and not sample_labels[key][label_type]:
        sample_labels.pop(key, None)



"""
Inter-process communication handling.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""

import pickle
import socket
import socketserver
import threading


class IPCServer(socketserver.TCPServer):
    """Server that listens for inter-process messages.

    Messages are exchanged as pickle-encoded objects.
    """

    timeout = 2

    def __init__(self, on_message):
        """Constructs (but does not start) a server.

        Args:
            on_message (function): callback that takes a single argument (any
                Python object sent by a client) and returns an object in
                response
        """
        super().__init__(
            ("localhost", 0), IPCRequestHandler, bind_and_activate=True
        )
        self.on_message = on_message
        self.__lock = threading.Lock()
        self.__in_serve_forever = False

    def stop(self):
        """Stops and shuts down the server.

        Interrupts serve_forever() if necessary."""
        with self.__lock:
            if self.__in_serve_forever:
                self.shutdown()
        self.server_close()

    def serve_forever(self):
        """Handles one request at a time until the server shuts down.

        This is a simple wrapper around TCPServer.serve_forever that keeps
        track of whether it was called.
        """
        try:
            with self.__lock:
                self.__in_serve_forever = True
            super().serve_forever()
        finally:
            with self.__lock:
                self.__in_serve_forever = False

    @property
    def port(self):
        return self.server_address[1]

    @classmethod
    def run_in_background(cls, on_message):
        """Convenience wrapper that creates a new server instance and calls
        serve_forever in a background thread.

        The server will be listening before this function returns.

        Returns:
            the IPCServer instance
        """
        server = cls(on_message)
        thread = threading.Thread(target=server.serve_forever)
        thread.daemon = True
        thread.start()
        return server

    # backport for Python 3.5 - these were added in Python 3.6
    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.server_close()


class IPCRequestHandler(socketserver.StreamRequestHandler):
    def handle(self):
        try:
            message = pickle.load(self.rfile)
            reply = self.server.on_message(message)
            self.send_reply(reply)
        except Exception as e:
            self.send_reply(e)

    def send_reply(self, reply):
        pickle.dump(reply, self.wfile)


def send_request(port, message):
    """Send a request to an IPCServer.

    Args:
        port (int): port to connect to
        message (any type)

    Returns:
        response (any type)
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(("localhost", port))
    pickle.dump(message, s.makefile("wb"))
    return pickle.load(s.makefile("rb"))



"""
FiftyOne Server filter inputs

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import strawberry as gql
from strawberry.schema_directive import Location
import typing as t


@gql.schema_directive(locations=[Location.INPUT_OBJECT])
class OneOf:
    pass


@gql.input
class GroupElementFilter:
    id: t.Optional[str] = None
    slice: t.Optional[str] = None


@gql.input  # oneof not working
class SampleFilter:
    id: t.Optional[str] = None
    group: t.Optional[GroupElementFilter] = None



"""
FiftyOne Server interfaces

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import typing as t

import strawberry as gql


@gql.interface
class StageParameter:
    _cls: str
    name: str
    kind: str


@gql.interface
class Stage:
    _cls: str
    kwargs: t.List[StageParameter]



"""
FiftyOne v0.7.3 revision.

| Copyright 2017-2022, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""


def up(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)
    if "evaluations" not in dataset_dict:
        dataset_dict["evaluations"] = {}

    db.datasets.replace_one(match_d, dataset_dict)


def down(db, dataset_name):
    match_d = {"name": dataset_name}
    dataset_dict = db.datasets.find_one(match_d)
    dataset_dict.pop("evaluations", None)
    db.datasets.replace_one(match_d, dataset_dict)
